#ifndef __FIXED_ARRAY_H
#define __FIXED_ARRAY_H

#include <boost/utility.hpp>

template <class Value_T>
class FixedArray
  {
public:
    typedef Value_T ValueType;
    typedef FixedArray<ValueType> thistype;
private:
    ValueType *mArr;
    unsigned mSize;

    void copy(const thistype &cpy)
      {
        mSize = cpy.mSize;
        if (mSize && cpy.mArr)
          {
            mArr = new ValueType[mSize];
            for (int i=0; i<mSize; ++i)
              mArr[i] = cpy.mArr[i];
          }
        else
          mArr = 0;
      }

public:
    FixedArray() : mArr(0), mSize(0) { }
    explicit FixedArray(unsigned size) : mArr(new ValueType[size]), mSize(size) { }
    FixedArray(const thistype &cpy)
      { copy(cpy); }
    void clear()
      {
        delete[] mArr;
        mSize=0;
        mArr=0;
      }
    ~FixedArray()
      { clear(); }
    const thistype & operator = (const thistype &cpy)
      { clear(); copy(cpy); return *this; }
    bool IsInitialized()
      {
        assert(mArr!=0 || mSize==0);
        return mArr!=0;
      }
    void SetSize(unsigned size)
      {
        assert(!IsInitialized());
        mArr=new ValueType[size];
        mSize=size;
      }
    typedef ValueType *iterator;
    typedef const ValueType *const_iterator;
    iterator begin() 
                    { assert(IsInitialized()); return mArr; }
    iterator end()   
                    { assert(IsInitialized()); return mArr+mSize; }
    const_iterator begin() const 
                    { assert(IsInitialized()); return mArr; }
    const_iterator end()   const 
                    { assert(IsInitialized()); return mArr+mSize; }
    ValueType & operator[](int index)
      {
        assert(IsInitialized());
        assert(index>=0);
        assert(index<mSize);
        return mArr[index];
      }
    bool empty()
      { return mSize==0; }
    int size()
      { return mSize; }
    void swap(thistype &swapwith)
      {
        swap(mArr, swapwith.mArr);
        swap(mSize, swapwith.mSize);
      }
  };

namespace std {
template<class Value_T>
void swap(FixedArray<Value_T> &a, FixedArray<Value_T> &b)
  {
    a.swap(b);
  }
}
#endif /* __FIXED_ARRAY_H */