#include "profiler.h"
#include <assert.h>
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>

class ProfilerImpl : public Profiler
  {
protected:
   
    static LARGE_INTEGER get_time_reading_int();

  	LARGE_INTEGER mBaseTime;
    double mStartTime;
    Profiling_Int64 mStartIntegerTimestamp;

    double get_time_double() 
      {
        LARGE_INTEGER time = get_time_reading_int();
	    time.QuadPart = time.QuadPart - mBaseTime.QuadPart;
        return (double)time.QuadPart * 0.001;
      }

public:
    ProfilerImpl()
      {
      	mBaseTime = get_time_reading_int();
      }

    virtual void Start();
    virtual void Done();
  };

LARGE_INTEGER ProfilerImpl :: get_time_reading_int()
  {
    LARGE_INTEGER freq;
    LARGE_INTEGER time;

    BOOL ok = QueryPerformanceFrequency(&freq);
    assert(ok == TRUE);

    freq.QuadPart = freq.QuadPart / 1000;

    ok = QueryPerformanceCounter(&time);
    assert(ok == TRUE);

    time.QuadPart = time.QuadPart / freq.QuadPart;

  	return time;
  }

void ProfilerImpl::Start()
  {
    for (int i = 0; i < MAX_PROFILING_ZONES; i++) 
      {
        Program_Zone *zone = Profiling::zone_pointers_by_index[i];
        if (zone == NULL) break;

        zone->total_self_ticks = 0;
        zone->total_hier_ticks = 0;
        zone->total_entry_count = 0;
      }

    mStartTime=get_time_double();
    os_fast_get_integer_timestamp(&mStartIntegerTimestamp);
  }

void ProfilerImpl::Done()
  {
    double totaltime=get_time_double();
    Profiling_Int64 totaltimestamps;
    os_fast_get_integer_timestamp(&totaltimestamps);
    totaltime -= mStartTime;
    totaltimestamps -= mStartIntegerTimestamp;
    double timestamps_to_seconds = totaltime/totaltimestamps;

    mReport.clear();

    for (int i = 0; i < MAX_PROFILING_ZONES; i++) 
      {
        Program_Zone *zone = Profiling::zone_pointers_by_index[i];
        if (zone == NULL) break;

        mReport.push_back();
        mReport.back().mName = zone->name;
        mReport.back().mSelfTime = zone->total_self_ticks*
                                    timestamps_to_seconds;
        mReport.back().mWholeTime = zone->total_hier_ticks*
                                    timestamps_to_seconds;
        mReport.back().mNumExecutions = zone->total_entry_count;
      }
  }

std::auto_ptr<Profiler> Profiler::Create()
  {
    return new ProfilerImpl;
  }
