#include "root2d.h"
#include "findcntr.h"
#include <hash_set>
#include "merge2d.h"

// should replace STPoint  with SmallMatrix<ScalarType, 1, 2>

namespace Root2D
  {
    
    void Cubic2dContourFinder 
    ::OutputVertex(std::list<ContourType> & outto, 
                            const STLocation &st)
      {
        assert(Matrix::Abs(st(0,0))<1.e50);
        assert(Matrix::Abs(st(0,1))<1.e50);
        /*
        if (st(0,0)<0 || st(0,1)<0 || st(0,0)+st(0,1)>1)
          {
            STLocation st2;
            if (st(0,0)<0)
              st2(0,0)=0;
            else
              st2(0,0)=st(0,0);
            if (st(0,1)<0)
              st2(0,1)=0;
            else
              st2(0,1)=st(0,1);
            ScalarType sum=st2(0,0)+st2(0,1);
            if (sum>1)
              Matrix::Scale(st2, 1/sum);
            outto.back().mControl.push_back(mCubic.STToDomain(st2));
          }
        else
        */
          outto.back().mControl.push_back(mCubic.STToDomain(st));
      }


    
    Cubic2dContourFinder ::STLocation
    Cubic2dContourFinder 
    ::PushLineSegmentResult(std::list<ContourType> & outto, 
                            const STLocation &st, ScalarType s, ScalarType t)
      {
        STLocation st2;
        st2(0,0)=s;
        st2(0,1)=t;
        STLocation temp(st);
        Matrix::AddScaledDiff(temp, 1./3, st2, st);
        OutputVertex(outto, temp);
        Matrix::AddScaledDiff(temp, 1./3, st2, st);
        OutputVertex(outto, temp);
        OutputVertex(outto, st2);
        return st2;
      }

    
    void Cubic2dContourFinder 
    ::OutputStraightLineSegment(std::list<ContourType> & outto,
                                const STLocation &st1, const STLocation &st2, 
                                int side1, int side2)
      {
        outto.push_back();
        OutputVertex(outto, st1);
        outto.back().mEndingSides[0]=side1;
        outto.back().mEndingSides[1]=side2;
        PushLineSegmentResult(outto, st1, st2(0,0), st2(0,1));
      }

    
    void Cubic2dContourFinder 
    ::FactorOutZeroSides()
      {
        for (int side=0; side<3; ++side)
          {
            mZeroSides[side]=false;
            while (1)
              {
                Cubic1dUnit b;
                mCubic.CubicFromSide(side, b);
                // if side is identically zero, have to rule out that side
                if (BezierIsZero(b, mEpsilonR))
                  {
                    mZeroSides[side]=true;
                    TriangularQuadraticUnit quadratic;
                    if (side==0)
                      {
                        quadratic.mY[0]= mCubic.mY[4]*3;
                        quadratic.mY[1]= mCubic.mY[5]*3;
                        quadratic.mY[2]= mCubic.mY[6]*3;
                        quadratic.mY[3]= mCubic.mY[7]*1.5;
                        quadratic.mY[4]= mCubic.mY[8]*1.5;
                        quadratic.mY[5]= mCubic.mY[9];
                      }
                    else if (side==1)
                      {
                        quadratic.mY[0]= mCubic.mY[0];
                        quadratic.mY[1]= mCubic.mY[1]*1.5;
                        quadratic.mY[2]= mCubic.mY[2]*3;
                        quadratic.mY[3]= mCubic.mY[4]*1.5;
                        quadratic.mY[4]= mCubic.mY[5]*3;
                        quadratic.mY[5]= mCubic.mY[7]*3;
                      }
                    else /* side==2 */
                      {
                        quadratic.mY[0]= mCubic.mY[1]*3;
                        quadratic.mY[1]= mCubic.mY[2]*1.5;
                        quadratic.mY[2]= mCubic.mY[3];
                        quadratic.mY[3]= mCubic.mY[5]*3;
                        quadratic.mY[4]= mCubic.mY[6]*1.5;
                        quadratic.mY[5]= mCubic.mY[8]*3;
                      }
                    Root2D::DegreeRaise(quadratic, mCubic);
                  }
                else
                  break;
              }
          }
      }
  
    // find the roots on each side
    
    void Cubic2dContourFinder 
    ::FindRootsOnEachSide()
      {
        for (int side=0; side<3; ++side)
          {
            Cubic1dUnit b;
            assert(mRootsOnSides[side].size()==0);
            mCubic.CubicFromSide(side, b);
#ifndef N_ROOTFINDER
            Root1D::Solve_Quadratic1D(mRootsOnSides[side], b, 
                                      mEpsilon, mEpsilonR);
#else
            Root1D::Solve_N_1D(mRootsOnSides[side], b, 
                                      mEpsilon, mEpsilonR);
#endif
            assert(mRootsOnSides[side].size()<=3);
          }
      }

        // decide on a suitable orientation for the triangle
    
    void Cubic2dContourFinder 
    ::IncrementGoodSide()
      { // rotate mCubic (and rotate roots on side)
        ++mGoodSide;
        mCubic.RotateClockwise();
        // Before   After
        //  2|\1    0|\2
        //    0       1
        mRootsOnSides[0].swap(mRootsOnSides[2]);  // 0|\1
                                                  //   2
        mRootsOnSides[0].swap(mRootsOnSides[1]);  // 0|\2
                                                  //   1
      }

    int Cubic2dContourFinder 
    ::DetermineGoodSide()
      {
        while (mGoodSide<3)
          {
            mBadSides[mGoodSide]=false;
            mCubicFinder.Initialize(mCubic);
            // test if orientation is good:
            // find cubic corresponding to each horizontal line 
            // through a root on each side
            // if cubic is identically 0 (see test in RootsInInterval)
            // then this orientation is bad
            // (note: this only happens if two sides have roots with the 
            //  same "t", so I only check if two sides have close roots)
            for (RootIter iter=mRootsOnSides[1].begin();
                 iter != mRootsOnSides[1].end(); ++iter)
              {
                ScalarType t=(*iter);
                // (careful! don't be fooled by t=1 corner 
                //  being in a contour)
                if (t<1-mEpsilon)
                  {
                    // look for a corresponding root on side 2
                    for (RootIter iter2=mRootsOnSides[2].begin();
                         iter2 != mRootsOnSides[2].end(); ++iter2)
                      {
                        ScalarType neart=1-(*iter2);
                        ScalarType r=Matrix::Abs(t-neart);
                        if (r<2*mEpsilon)
                          {
                            // four points determine a cubic -- we know two 
                            // endpoints are zero, so test two interior points
                            ScalarType st[2];
                            st[1] = (t+neart)*.5;
                            st[0] = (1-st[1])*(1./3);
                            if (NearAZero(mCubic, st, mEpsilon, mEpsilonR))
                              {
                                st[0] = (1-st[1])*(2./3);
                                if (NearAZero(mCubic, st, mEpsilon, mEpsilonR))
                                  { // four points in a horizontal are zero
                                    // so this orientation is bad
                                    mBadSides[mGoodSide]=true;
                                    mBadT[mGoodSide]=(t+neart)*.5;
                                    break;
                                  }
                              }
                          }
                      }
                    if (mBadSides[mGoodSide])
                      break;
                  }
              }

            if (mBadSides[mGoodSide])
              IncrementGoodSide();
            else
              // else: orientation is good, we're done
              return 0;
          }

        // didn't find a good side
        if (mBadSides[0] && mBadSides[1] && mBadSides[2])
          ThreeStraightLinesCase();
        else
          // something else went wrong, need a better epsilon
          return 2;

        return 1;
      }

    
    void Cubic2dContourFinder 
    ::ThreeStraightLinesCase()
      {
        // did not find a good orientation
        // actually, this case only happens when the function
        // has contours = 3 straight lines parallel to the 3
        // sides of the triangle
        // using "mBadT" above we can figure out the positions
        // of those three lines
        // Only one question remains: 
        //
        //   \ |            \   |       ?
        //    \|             \  |       ?
        //                    | \       ?
        //      |\    OR      |  \      ?
        //      | \        ___|   \___  ?
        //   ___ ~~ __         |~~\     ?
        //     |    \          |   \    ?
        //     |     \         |    \   ?
        //        or maybe:
        //   \ |            \   |       ?
        //    \|             \  |       ?
        //                    | \       ?
        //     |\    OR       |  \      ?
        //     | \         ___ \  \___  ?
        //   __//~~ __        \ ~~\     ?
        //      |   \          |   \    ?
        //      |    \         |    \   ?

        //for (int i=0; i<3; ++i)
        //  if (mBadT[i]<mEpsilon)
        //    mBadT[i]=0;
        //  else if (mBadT[i]>1-mEpsilon)
        //    mBadT[i]=1;

        if (mBadT[0]+mBadT[1]+mBadT[2] < 1)
          { // Using:
            //     \   |                 .
            //      \  |                 .
            //       | \                 .
            //       |  \                .
            //    ___|   \___   mBadT[0] .
            //        |~~\               .
            //        |   \              .
            //        |    \             .
            //   mBadT[2] mBadT[1]

            assert(mBadT[0]>mEpsilonR);
            assert(mBadT[1]>mEpsilonR);
            assert(mBadT[2]>mEpsilonR);

            STLocation st;
            // Bottom segment -------
            (*mResults).push_back();
            // output mBadT[2],0 
            st(0,0)=mBadT[2];
            st(0,1)=0;
            OutputVertex(*mResults, st);

            (*mResults).back().mEndingSides[0]=mBadT[2]>mEpsilonR ? 0 : 3;
            (*mResults).back().mEndingSides[1]=mBadT[1]>mEpsilonR ? 0 : 4;
            if (mBadT[0]>mEpsilonR)
              { // output -> mBadT[2],mBadT[0]
                st=PushLineSegmentResult(*mResults, st, mBadT[2], mBadT[0]);
              }

            // output -> 1-mBadT[0]-mBadT[1],mBadT[0]
            st=PushLineSegmentResult(*mResults, st, 1-mBadT[0]-mBadT[1], mBadT[0]);

            if (mBadT[0]>mEpsilonR)
              // output -> 1-mBadT[1],0
              st=PushLineSegmentResult(*mResults, st, 1-mBadT[1], 0);

            // Right segment -------
            (*mResults).push_back();
            // output 1-mBadT[0], mBadT[0]
            st(0,0)=1-mBadT[0];
            st(0,1)=mBadT[0];
            OutputVertex(*mResults, st);

            (*mResults).back().mEndingSides[0]=mBadT[0]>mEpsilonR ? 1 : 4;
            (*mResults).back().mEndingSides[1]=mBadT[2]>mEpsilonR ? 1 : 5;
            if (mBadT[1]>mEpsilonR)
              { // output -> 1-mBadT[0]-mBadT[1],mBadT[0]
                st=PushLineSegmentResult(*mResults, st, 1-mBadT[0]-mBadT[1], mBadT[0]);
              }

            // output -> mBadT[2], 1-mBadT[2]-mBadT[1]
            st=PushLineSegmentResult(*mResults, st, mBadT[2], 1-mBadT[2]-mBadT[1]);

            if (mBadT[1]>mEpsilonR)
              //  -> mBadT[1],1-mBadT[1]
              st=PushLineSegmentResult(*mResults, st, mBadT[2], 1-mBadT[2]);

            // Left segment -------
            (*mResults).push_back();
            // output 0, 1-mBadT[1]
            st(0,0)=0;
            st(0,1)=1-mBadT[1];
            OutputVertex(*mResults, st);

            (*mResults).back().mEndingSides[0]=mBadT[1]>mEpsilonR ? 2 : 5;
            (*mResults).back().mEndingSides[1]=mBadT[2]>mEpsilonR ? 2 : 3;
            if (mBadT[2]>mEpsilonR)
              { // output -> mBadT[2], 1-mBadT[2]-mBadT[1]
                st=PushLineSegmentResult(*mResults, st, mBadT[2], 1-mBadT[2]-mBadT[1]);
              }

            // output -> mBadT[2], mBadT[0]
            st=PushLineSegmentResult(*mResults, st, mBadT[2], mBadT[0]);

            if (mBadT[2]>mEpsilonR)
              //  -> 0, mBadT[0]
              st=PushLineSegmentResult(*mResults, st, 0, mBadT[0]);
          }
        else
          { // MANY MANY CASES!
            // Using:
            //  4      3          .
            //   \_____|          .
            //  5_8    7_2 mBadT[0]
            //    \    |          .
            //     \   |          .
            //      \ 6|          .
            //       | \          .
            //       0  1         .
            // mBadT[2] mBadT[1]  .

            // side numbering (these sides are merged and removed by
            //                  MergeContours)
            // \       |     .
            //  \10   9|     .
            //   \_____|     .
            // ___      ___  .
            //  11\    |8    .
            //     \   |     .
            //      \  |     .
            //      7| \6    .
            //       |  \    .

            STLocation st[9];
            bool inside[3];
            bool equal[3];
            st[0](0,0)=mBadT[2];    st[0](0,1)=0;
            st[1](0,0)=1-mBadT[1];  st[1](0,1)=0;
            st[2](0,0)=1-mBadT[0];  st[2](0,1)=mBadT[0];
            st[3](0,0)=mBadT[2];    st[3](0,1)=1-mBadT[2];
            st[4](0,0)=0;           st[4](0,1)=1-mBadT[1];
            st[5](0,0)=0;           st[5](0,1)=mBadT[0];

            st[6](0,0)=mBadT[2];            st[6](0,1)=1-mBadT[1]-mBadT[2];
            st[7](0,0)=mBadT[2];            st[7](0,1)=mBadT[0];
            st[8](0,0)=1-mBadT[1]-mBadT[0]; st[8](0,1)=mBadT[0];

            inside[0] = st[6](0,1) >= 0;
            equal [0] = st[6](0,1) == 0;
            inside[1] = st[7](0,0) + st[7](0,1) <= 1;
            equal [1] = st[7](0,0) + st[7](0,1) == 1;
            inside[2] = st[8](0,0) >= 0;
            equal [2] = st[8](0,0) == 0;
            std::list<ContourType> inner;
            if (mBadT[0]+mBadT[1]+mBadT[2] > 1)
              {
                if (inside[0] && inside[1])
                  {
                    OutputStraightLineSegment(inner, st[6], st[7], 6, 8);
                  }
                if (inside[1] && inside[2])
                  {
                    OutputStraightLineSegment(inner, st[7], st[8], 9, 10);
                  }
                if (inside[2] && inside[0])
                  {
                    OutputStraightLineSegment(inner, st[8], st[6], 11, 7);
                  }
              }
            // verts 0&1
            if (inside[0])
              { // 06 16
                OutputStraightLineSegment(*mResults, st[0], st[6], 
                                          st[0](0,0)>0 ? 0 : 3,
                                          st[0](0,0)>0 ? 7 : 2);
                OutputStraightLineSegment(*mResults, st[1], st[6], 
                                          st[1](0,0)<1 ? 0 : 4,
                                          st[1](0,0)<1 ? 6 : 1);
              }
            else 
              {
                if (inside[1])
                  // 07
                  OutputStraightLineSegment(inner, st[0], st[7], 0, 8);
                else
                  // 03
                  OutputStraightLineSegment(*mResults, st[0], st[3], 0, 1);
                if (inside[2])
                  // 18
                  OutputStraightLineSegment(inner, st[1], st[8], 0, 11);
                else
                  // 14
                  OutputStraightLineSegment(*mResults, st[1], st[4], 0, 2);
              }
            // verts 2&3
            if (inside[1])
              { // 27 37
                OutputStraightLineSegment(*mResults, st[2], st[7], 
                                          st[2](0,1)>0 ? 1 : 4,
                                          st[2](0,1)>0 ? 8 : 0);
                OutputStraightLineSegment(*mResults, st[3], st[7],
                                          st[3](0,0)>0 ? 1 : 5,
                                          st[3](0,0)>0 ? 9 : 2);
              }
            else 
              {
                if (inside[2])
                  // 28
                  OutputStraightLineSegment(inner, st[2], st[8], 1, 10);
                else
                  // 25
                  OutputStraightLineSegment(*mResults, st[2], st[5], 1, 2);
                if (inside[0])
                  // 36
                  OutputStraightLineSegment(inner, st[3], st[6], 1, 6);
                // else 03 already output above
              }
            // verts 4&5
            if (inside[2])
              { // 48 58
                OutputStraightLineSegment(*mResults, st[4], st[8], 
                                          st[4](0,1)<1 ? 2 : 5,
                                          st[4](0,1)<1 ? 10 : 1);
                OutputStraightLineSegment(*mResults, st[5], st[8], 
                                          st[5](0,1)>0 ? 2 : 3,
                                          st[5](0,1)>0 ? 11 : 0);
              }
            else 
              {
                if (inside[0])
                  // 46
                  OutputStraightLineSegment(inner, st[4], st[6], 2, 7);
                // else 14 already drawn
                if (inside[1])
                  // 57
                  OutputStraightLineSegment(inner, st[5], st[7], 2, 9);
                // else 25 already drawn
              }
            int tomerge[6] = { 6, 7, 8, 9, 10, 11 };
            Merge2D::ConnectContours(*mResults, inner, mEpsilon, 6, tomerge);
          }
      }

       // compute gradient of mCubic in "s" and "t" direction
    
    void Cubic2dContourFinder 
    ::ComputeGradients()
      {
        mCubic.PartialDerivativeS(mGradient[0]);
        mCubic.PartialDerivativeT(mGradient[1]);
      }


            // Classifying these intersections and roots:
            // "StartAtEdge", "EndAtEdge", "Max", "Min", "Isolated", etc.
            // (at edge points, this is easy assuming the gradient is 
            // nonzero in the "s" direction and in the perpendicular-to-
            // edge direction)
            // (at corners, must have same condition of "edge start" 
            // "edge end" on both sides, otherwise just an isolated 
            // point of the contour)
            // (at interior points, could use second derivative, if 
            // nonzero, to classify as either "starting" or "ending" -- 
            // see equation 1.20 on page 8 of Grandine 1997:
            // think it translates to:
            //    f_t * f_ss < 0 => "starting")
            // There are several ways these test can be indeterminant:
            // Edge case:
            //    contour horizontal at edge => use f_t * f_ss <> 0
            //          to see if concave up or down, if 0, see critical
            //          point case below
            //    contour tangent to edge => another 2nd derivative test
            //          to see if inside or outside (probably can ignore
            //          in either case)
            //          if 0, maybe inflection point
            //    gradient 0 => bad news? see critical point case below
            // Interior case:
            //    gradient 0, "critical point case" => compute determinant
            //          of hessian matrix: positive implies isolated point,
            //          negative implies saddle point, contours cross
            //          0 implies more bad news
            //    f_ss = 0 but f_t != 0 (example? t+s^3 at 0,0; t alone
            //          would be a horizontal line already eliminated)
            //          probably can ignore this case, since just an
            //          inflection point
            // Grandine punts, says other cases not 
            // implemented, "future work"
            // We'll mark these as "Horizontal" or "Unknown"
            // Then as a second pass, we'll find what happens mEpsilon
            // above and below and use that data to resolve the
            // ambiguity

        // first, the roots on each side
    
    void Cubic2dContourFinder 
    ::InsertEdgePoints()
      {
        TransitionPointList * corner_line[3] = { 0, 0, 0 };
        bool corner_process[3] = { false, false, false };
        int corner_ind[3];
        int i;
        for (i=0; i<3; ++i)
          {
            for (RootIter rootiter=mRootsOnSides[i].begin();
                 rootiter!=mRootsOnSides[i].end(); ++rootiter)
              {
                ScalarType st[2];
                bool iscorner=false;
                int whichcorner;
                switch (i)
                  {
                    case 0:
                      st[0] = *rootiter;
                      st[1] = 0;
                      break;
                    case 1:
                      st[1] = *rootiter;
                      st[0] = 1-st[1];
                      break;
                    case 2:
                      st[1] = 1 - *rootiter;
                      st[0] = 0;
                      break;
                    default:
                      assert(0);
                  }
                if (*rootiter<mEpsilon)
                  {
                    iscorner=true;
                    whichcorner=i;
                  }
                else if (*rootiter>1-mEpsilon)
                  {
                    iscorner=true;
                    whichcorner=(i+1) % 3;
                  }
                TransitionPointList *tline;
                TransitionPoint *point;
                if (iscorner && corner_line[whichcorner])
                  // already counted this corner
                  {
                    if (! corner_process[whichcorner])
                      continue;
                    tline=0;
                    point=new TransitionPoint;
                  }
                else
                  {
                    tline=FindLine(st[1], mEpsilon*.5);
                    (*tline).push_back();
                    point=& (*tline).back();
                    if (iscorner)
                      {
                        corner_line[whichcorner]=tline;
                        corner_ind[whichcorner]=(*tline).size()-1;
                      }
                  }
                point->mS=st[0];
                point->mT=st[1];

                ScalarType secondpartials[2][2]; 
                mGradient[0].EvaluateAt(st, point->mNormal[0],
                                        secondpartials[0]);
                mGradient[1].EvaluateAt(st, point->mNormal[1],
                                        secondpartials[1]);
                bool zero[2];
                zero[0]=NearAZero(mGradient[0], st, mEpsilon, mEpsilonR);
                zero[1]=NearAZero(mGradient[1], st, mEpsilon, mEpsilonR);
                //zero[0] = Matrix::Abs(point->mNormal[0])<mEpsilon;
                //zero[1] = Matrix::Abs(point->mNormal[1])<mEpsilon;
                point->mTransition=ClassifyEdgePoint(point, 
                                            secondpartials, zero, i, iscorner);
                if (point->mTransition == DeletePoint && tline)
                  { 
                    /*
                    TransitionPointIter toremove = mTData[st[1]].mTrans.end();
                    --toremove;
                    mTData[st[1]].mTrans.erase(toremove);
                    if (mTData[st[1]].mTrans.empty())
                      mTData.erase(st[1]);
                    */
                    continue;
                  }
#ifdef CURVATURE
                if (point->mTransition != Unknown && 
                    point->mTransition != UnknownAtEdge)
                  {
                    point->mCurvature=ComputeCurvature(point->mNormal, 
                                                        secondpartials);
                  }
#endif

                point->mSide = i;

                if (tline==0)
                  {
                    TransitionPoint *corner=&*(corner_line[whichcorner]->begin()+
                                               corner_ind[whichcorner]);
                    point->mSide=corner->mSide;
                    if (point->mT<corner->mT)
                      ConnectEdgePoints(*point, *corner);
                    else
                      ConnectEdgePoints(*corner, *point);
                    if (corner->mTransition!=UnknownAtEdge)
                      corner->mTransition=DeletePoint;
                    else  // BOGUS
                      corner->mTransition=UnknownAtEdge;
                    delete point;
                  }
                else if (iscorner)
                  { // special handling of corner cases
                    point->mSide = whichcorner+3;
                    int otherside;
                    // whichcorner has which corner the root is in
                    if (whichcorner==i)
                      otherside=(whichcorner+2)%3;
                    else
                      otherside=whichcorner;
                    TransitionType othertransition=ClassifyEdgePoint(point, 
                           secondpartials, zero, otherside, true);
                    if (othertransition!=point->mTransition)
                      { // if they agree, we're happy
                        // otherwise we need to resolve cases...
                        if (TransitionPair(othertransition,
                              point->mTransition,StartAtEdge,EndAtEdge))
                          corner_process[whichcorner]=true;
                        else if (TransitionPair(othertransition,
                              point->mTransition,NoTransition,EndAtEdge))
                          point->mTransition=EndAtEdge;
                        else if (TransitionPair(othertransition,
                              point->mTransition,NoTransition,StartAtEdge))
                          point->mTransition=StartAtEdge;
                        else if (TransitionPair(othertransition,
                              point->mTransition,DeletePoint,EndAtEdge))
                          corner_process[whichcorner]=true;
                        else if (TransitionPair(othertransition,
                              point->mTransition,Min,StartAtEdge))
                          point->mTransition=StartAtEdge;
                        else if (TransitionPair(othertransition,
                              point->mTransition,Isolated,EndAtEdge))
                          {
                            point->mTransition=Isolated;
                            corner_process[whichcorner]=true;
                          }
                        else
                        // DO ALL CASES
                          assert(0 && "Haven't dealt with this case yet");
                      }
                    if (whichcorner==2)
                      {
                        if (point->mTransition==HorizontalAtEdge)
                          corner_process[whichcorner]=true;
                        else if (point->mTransition==StartAtEdge)
                          {
                            point->mTransition=Isolated;
                            corner_process[whichcorner]=true;
                          }
                        else if (point->mTransition==EndAtEdge)
                          if (point->mNormal[1]>0)
                            {
                              if (point->mNormal[0] < point->mNormal[1])
                                corner_process[whichcorner]=true;
                            }
                          else if (point->mNormal[1]<0)
                            {
                              if (point->mNormal[0] > point->mNormal[1])
                                corner_process[whichcorner]=true;
                            }
                      }
                    if (point->mTransition==UnknownAtEdge ||
                        point->mTransition==Isolated)
                      corner_process[whichcorner]=true;
                  }
              }
          }
      }

            // Now process internal transition points, "interiorpoints"
    
    void Cubic2dContourFinder 
    ::InsertInteriorPoints()
      {
        // first compute them, the "CommonIntersection"s
        typedef std::list<STPoint  > STPointList;
        STPointList interiorpoints;
        SimultaneousSolve(mCubic, mGradient[0], mEpsilon, mEpsilonR, 
                          interiorpoints);
        // then classify them and add them to mTData
        while (! interiorpoints.empty())
          {
            STPoint  intpoint = interiorpoints.back();
            interiorpoints.pop_back();

            TransitionPoint  *point=
              FindDuplicate(intpoint.st[0], intpoint.st[1], mEpsilon);
            // check to see if this point already in mTData somewhere
            // probably should only occur if s,t indicate along edge
            if (point)
              {
                if (point->mSide==-1)
                  {
                    TransitionPoint  point2;
                    AnalyzeInteriorPoint(intpoint.st, &point2);
                    if (point->mTransition!=point2.mTransition)
                      point->mTransition=Unknown;
                  }
              }
            else
              {
                TransitionPointList *tpl=FindLine(intpoint.st[1], mEpsilon/2);
                (*tpl).push_back();
                point=&(*tpl).back();
                point->mS=intpoint.st[0];
                point->mT=intpoint.st[1];

                AnalyzeInteriorPoint(intpoint.st, point);
              }
          }
      }


    void Cubic2dContourFinder 
    ::RemoveDeletedPoints()
      {
        TDataIter tline;
        for (tline = mTData.begin(); tline != mTData.end() ; ++tline)
          {
        top:
            TransitionPointIter tliter = (*tline).second.mTrans.begin();
            while (tliter!=(*tline).second.mTrans.end())
              {
                if ((*tliter).mTransition==DeletePoint)
                  {
                    tliter=(*tline).second.mTrans.erase(tliter);
                    if ((*tline).second.mTrans.empty())
                      {
                        TDataIter toerase=tline;
                        ++tline;
                        mTData.erase(toerase);
                        if (tline != mTData.end())
                          goto top;
                        return;
                      }
                  }
                else
                  ++tliter;
              }
          }
      }
    

    void Cubic2dContourFinder 
    ::FindRootsAtSameT()
      {
        // compute zeros for each corresponding line of constant "t"
        TDataIter tline;
        for (tline = mTData.begin(); tline != mTData.end() ; ++tline)
          {
            ScalarType t = (*tline).first;
            Root1D::BezierInterval  b;
            mCubicFinder.CubicFromT(t, b);
            RootList roots;
#ifndef N_ROOTFINDER
            Root1D::Solve_Quadratic(roots, b, mEpsilon/2, mEpsilonR/4);
#else
            Root1D::Solve_N_P(roots, b, mEpsilon/2, mEpsilonR/4);
#endif
            assert(roots.size()<=3);
            // compare roots with (*tline).second
            std::sort((*tline).second.mTrans.begin(), (*tline).second.mTrans.end());
            // should no longer be necessary since Solve_N_P return roots
            // in increasing order
            roots.sort();
            TransitionPointIter tliter = (*tline).second.mTrans.begin();
            RootIter riter = roots.begin();
            // reasonable:
            // roots missing a root from (*tline).second
            // roots having 2-3 roots where (*tline).second has one
            // should add roots in roots missing from (*tline).second
            while (riter != roots.end())
              {
                if (tliter==(*tline).second.mTrans.end() || 
                    (*riter) < (*tliter).mS - mEpsilon)
                  { // add anything missing from *tline as "NoTransition"
                    // (compute the normal vector, as well)
                    TransitionPoint  toadd;
                    toadd.mS=(*riter);
                    toadd.mT=t;
                    toadd.mTransition=NoTransition;
                    STPoint  st;
                    st.st[0]=(*riter);
                    st.st[1]=t;
#ifdef CURVATURE
                    ScalarType secondpartials[2][2];
                    mGradient[0].EvaluateAt(st.st, toadd.mNormal[0],
                                           secondpartials[0]);
                    mGradient[1].EvaluateAt(st.st, toadd.mNormal[1],
                                           secondpartials[1]);
                    assert(Matrix::Abs(toadd.mNormal[0])>mEpsilonR);
                    toadd.mCurvature=ComputeCurvature(toadd.mNormal, 
                                           secondpartials);
#else
                    toadd.mNormal[0]=mGradient[0].EvaluateAt(st.st);
                    toadd.mNormal[1]=mGradient[1].EvaluateAt(st.st);
                    assert(Matrix::Abs(toadd.mNormal[0])>mEpsilonR);
#endif
                    toadd.mSide = -1;
                    tliter = (*tline).second.mTrans.insert(tliter, toadd);
                    ++tliter;
                    ++riter;
                  }
                else
                  {
                    while (riter!=roots.end() && 
                           (*riter) <= (*tliter).mS + mEpsilon)
                      ++riter;
                    ++tliter;
                  }
              }
          }
      }

    
    void Cubic2dContourFinder 
    ::RemoveRedundantInterior()
      {
        TDataIter tline;
        for (tline = mTData.begin(); tline != mTData.end() ; ++tline)
          {
            TransitionPointIter tliter = (*tline).second.mTrans.begin();
            assert(tliter != (*tline).second.mTrans.end());

            TransitionPointIter tlnext = tliter;
            ++tlnext;
            for (; tlnext != (*tline).second.mTrans.end(); tliter=tlnext, ++tlnext)
              {
                TransitionPointType *minmax;
                TransitionPointType *notrans;
                if ((*tliter).mTransition == NoTransition &&
                    (*tlnext).mTransition != NoTransition)
                  {
                    notrans = & (*tliter);
                    minmax = & (*tlnext);
                  }
                else if ((*tlnext).mTransition == NoTransition &&
                         (*tliter).mTransition != NoTransition)
                  {
                    notrans = & (*tlnext);
                    minmax = & (*tliter);
                  }
                else
                  continue;
#ifdef CURVATURE
                if (minmax->mTransition==Min || minmax->mTransition==Max)
                  {
                    // compute the radius of curvature for both points
                    ScalarType rad_nt=1/notrans->mCurvature;
                    ScalarType rad_mm=1/minmax ->mCurvature;
                    ScalarType abs_nt=Matrix::Abs(rad_nt);
                    ScalarType abs_mm=Matrix::Abs(rad_mm);
                    // if difference between radii is > err, continue
#pragma fixme("is err value justified?")
                    ScalarType diff=sqrt(mEpsilon)*(abs_nt+abs_mm+1);
                    diff -= Matrix::Abs(abs_nt-abs_mm);
                    if (diff<0)
                      continue;

                    // compute the centers of the osculating circles
                    STLocation centers[2];
                    centers[0](0,0)=minmax->mS;
                    centers[0](0,1)=0;
                    FindCircleCenter(centers[0], rad_mm, minmax->mNormal);

                    centers[1](0,0)=notrans->mS;
                    centers[1](0,1)=0;
                    FindCircleCenter(centers[1], rad_nt, notrans->mNormal);

                    // compute distance between circle centers plus
                    // difference between radii
                    // if greater than epsilon, continue
                    if (diff<Matrix::Length(centers[0], centers[1]))
                      continue;
                  }
                else
#endif
                  {
                    ScalarType invslope=notrans->mNormal[1]/notrans->mNormal[0];
                    ScalarType deltas=notrans->mS-minmax->mS;
#ifndef CURVATURE
                    // if signs are wrong, continue
                    if ((minmax->mTransition==Min && 
                         SignOfProduct(invslope, deltas)==Negative) ||
                        (minmax->mTransition==Max && 
                         SignOfProduct(invslope, deltas)==Positive))
                      continue;
#endif
                    
                    // if deltas greater than epsilon*invslope, continue
                    if (Matrix::Abs(deltas)>Matrix::Abs(mEpsilon*invslope))
                      continue;
                  }

                // if we got here (no "continue"), then we've determined
                // that we should erase notrans

                // before erasing notrans, set tlnext to first of next pair
                // that we will test in next iteration of loop
                TransitionPointIter toerase;
                if ((*tlnext).mTransition == NoTransition)
                  toerase=tlnext;
                else
                  toerase=tliter;
                tlnext=(*tline).second.mTrans.erase(toerase);
                if (tlnext!=(*tline).second.mTrans.begin())
                  --tlnext;
              }
          }
      }

    void Cubic2dContourFinder 
    ::ConnectEdgePoints(const TransitionPoint &a, const TransitionPoint &b)
      {
        SplinePoint endpts[2];
        const SplinePoint *endptrs[2] = { & endpts[0], & endpts[1] };
        STLocation controlpts[2];
        const TransitionPoint *tp[2] = { &a, &b };

        // set endpts[...]
        int i;
        for (i=0; i<2; ++i)
          {
            endpts[i].mST(0,0)=tp[i]->mS;
            endpts[i].mST(0,1)=tp[i]->mT;
          }

        STLocation direct;
        Matrix::Subtract(direct, endpts[1].mST, endpts[0].mST);

        for (i=0; i<2; ++i)
          {
            if (-tp[i]->mNormal[1]*direct(0,0)+tp[i]->mNormal[0]*direct(0,1) > 0)
              {
                endpts[i].mTangent(0,0) = -tp[i]->mNormal[1];
                endpts[i].mTangent(0,1) = tp[i]->mNormal[0];
                // normal to right of tangent, 
                // so flip sign of curvature
                endpts[i].mCurvature = -tp[i]->mCurvature;
              }
            else
              {
                endpts[i].mTangent(0,0) = tp[i]->mNormal[1];
                endpts[i].mTangent(0,1) = -tp[i]->mNormal[0];
                endpts[i].mCurvature = tp[i]->mCurvature;
              }
            Matrix::Scale(endpts[i].mTangent, 
                          1/Matrix::TwoNorm(endpts[i].mTangent));
          }

        if (a.mTransition==UnknownAtEdge)
          if (b.mTransition==UnknownAtEdge)
            {
              OutputStraightLineSegment(*mResults, endpts[0].mST, 
                                        endpts[1].mST, 
                                        RotateSide(mGoodSide, a.mSide), 
                                        RotateSide(mGoodSide, b.mSide));
              return;
            }
          else
            UnknownToKnown(endpts[0], endpts[1], controlpts[0], controlpts[1]);
        else if (b.mTransition==UnknownAtEdge)
          UnknownToKnown(endpts[1], endpts[0], controlpts[1], controlpts[0]);
        else
          SimpleJoin(endptrs, controlpts);

        (*mResults).push_back();
        OutputVertex(*mResults, endpts[0].mST);
        OutputVertex(*mResults, controlpts[0]);
        OutputVertex(*mResults, controlpts[1]);
        OutputVertex(*mResults, endpts[1].mST);

        (*mResults).back().mEndingSides[0]=RotateSide(mGoodSide, a.mSide);
        (*mResults).back().mEndingSides[1]=RotateSide(mGoodSide, b.mSide);
      }

            // Initialize panel data structure and
            // compute a regular line in the middle of each panel
    void Cubic2dContourFinder 
    ::InitializePanels()
      {
        TDataIter tline, tline2;

#ifndef NDEBUG
        for (tline = mTData.begin(); tline != mTData.end() ; ++tline)
          {
            ScalarType t = (*tline).first;
            assert((*tline).second.mLastT >= t);
            assert((*tline).second.mLastT < t+mEpsilon);
            for (TransitionPointIter s=(*tline).second.mTrans.begin();
                 s!=(*tline).second.mTrans.end(); ++s)
              assert(Matrix::Abs((*s).mT-t)<4*mEpsilon);
          }
#endif

        for (tline = mTData.begin(), tline2 = tline, ++tline2; 
              tline2 != mTData.end() ; tline=tline2, ++tline2)
          if ((*tline).second.mLastT + 4*mEpsilon < (*tline2).first)
            {
              mPanels.push_back();
              mPanels.back().beginend[0]=tline;
              mPanels.back().beginend[1]=tline2;
              ScalarType midt = ((*tline).first + (*tline2).first)/2;
              RegularLine(midt, mEpsilon, mTData[midt]);
              mPanels.back().numlines=mTData[midt].mTrans.size();
              assert(mPanels.back().numlines<=3); // cubic function!
            }
          else
            {
              // aggregate all roots in the thin panel together
              assert((*tline).second.mLastT<(*tline2).first);
              (*tline).second.mLastT=(*tline2).first;

              TransitionPointIter s1=(*tline).second.mTrans.begin(), 
                s2=(*tline2).second.mTrans.begin();
              while (s1!=(*tline).second.mTrans.end() && 
                     s2!=(*tline2).second.mTrans.end())
                {
                  if (Matrix::Abs((*s1).mS-(*s2).mS) < 4*mEpsilon)
                    {
                      if (!MergePoints(*s1, *s2))
                        { // Should be connected instead of merged!
                          ConnectEdgePoints(*s1, *s2);
                          // remove from list
                          if ((*s1).mTransition==UnknownAtEdge)
                            ++s2; // skip (*s2)
                          else 
                            { // remove (*s1)
                              s1 = (*tline).second.mTrans.erase(s1);
                              if ((*s2).mTransition!=UnknownAtEdge)
                                // and skip (*s2)
                                ++s2;
                            }
                        }
                      else
                        {
                          ++s1;
                          ++s2;
                        }
                    }
                  else if ((*s1).mS < (*s2).mS)
                    {
                      ++s1;
                    }
                  else
                    {
                      // insert *s2 into (*tline).second before s1
                      s1=(*tline).second.mTrans.insert(s1, *s2);
                      ++s1;
                      ++s2;
                    }
                }
              while (s2!=(*tline2).second.mTrans.end())
                {
                  // insert *s2 into (*tline).second before s1
                  assert(s1==(*tline).second.mTrans.end());
                  s1=(*tline).second.mTrans.insert(s1, *s2);
                  ++s1;
                  ++s2;
                }

              mTData.erase(tline2);
              tline2=tline;
            }
      }

        // single row of roots
        // better be just isolated points
    
    bool Cubic2dContourFinder 
    ::NoPanelsCase()
      {
        for (TDataIter tline = mTData.begin();
             tline != mTData.end();
             ++tline)
          {
            STLocation st;
            for (TransitionPointIter tliter = (*tline).second.mTrans.begin();
                 tliter != (*tline).second.mTrans.end(); ++tliter)
              {
                if ((*tliter).mTransition!=Isolated &&
                    (*tliter).mTransition!=UnknownAtEdge)
                  return false;
                assert((*tline).first>1-mEpsilon || 
                       (*tliter).mTransition==Isolated ||
                       (*tliter).mTransition==UnknownAtEdge);
                if ((*tliter).mTransition==UnknownAtEdge)
                  if ((*tline).first<(*tline).second.mLastT)
                    return false;
                  else
                    continue;
                (*mResults).push_back();
                (*mResults).back().mEndingSides[0]=(*tliter).mSide;
                (*mResults).back().mEndingSides[1]=(*tliter).mSide;
                st(0,0)=(*tliter).mS;
                st(0,1)=(*tliter).mT;
                OutputVertex(*mResults, st);
                OutputVertex(*mResults, st);
                OutputVertex(*mResults, st);
                OutputVertex(*mResults, st);
              }
          }
        return true;
      }

        // --- if have Horizontal, HorizontalAtEdge, Unknown, or 
        // UnknownAtEdge ---
        // need to find the roots and gradients above and below, resolve
        // cases based on which side of transition point intersections 
        // occur
    
    void Cubic2dContourFinder 
    ::ComputeGuideLines()
      {
        for (PanelIter curpanel=mPanels.begin(); curpanel!=mPanels.end();
             ++curpanel)
          {
            TDataIter panelcenter=(*curpanel).beginend[0];
            ++panelcenter;
#ifndef NDEBUG
            TDataIter beforepanelend=(*curpanel).beginend[1];
            --beforepanelend;
            assert(panelcenter==beforepanelend);
#endif

            for (int i=0; i<2; ++i)
              {
                TDataIter tline = (*curpanel).beginend[i];
                bool guideline=false;
                for (TransitionPointIter tliter=(*tline).second.mTrans.begin();
                     tliter != (*tline).second.mTrans.end(); ++tliter)
                  {
                    TransitionType transition=(*tliter).mTransition;
                    if (transition==HorizontalAtEdge || 
                        transition==Horizontal       ||
                        transition==UnknownAtEdge    ||
                        transition==Unknown)
                      {
                        guideline=true;
                        break;
                      }
                  }
                if (guideline)
                  {
                    ScalarType t;
                    if (i)
                      t = (*tline).first - 2*mEpsilon;
                    else
                      t = (*tline).second.mLastT + 2*mEpsilon;
                    ScalarType smalleps = mEpsilon;
                    TDataIter guidelineptr;
                    // check that the number of roots matches
                    // the regular line in the center of this panel
                    do
                      {
                        mTData[t].mTrans.clear();
                        RegularLine(t, smalleps, mTData[t]);
                        smalleps /= 2;
                      } while (smalleps>mEpsilon/WorstSmallEpsilon && 
                               mTData[t].mTrans.size()!=
                                  (*panelcenter).second.mTrans.size());
                    assert(mTData[t].mTrans.size()==
                                  (*panelcenter).second.mTrans.size());
                    (*curpanel).guidelines[i] = mTData.find(t);
                  }
                else
                  (*curpanel).guidelines[i] = mTData.end() ;
              }
          }
      }

        // set mBeforeAfter for transition points on panel boundaries
    
    bool Cubic2dContourFinder 
    ::SetBeforeAfter()
      {
        TransitionPointIter tliter;
        
        { // first make sure top and bottom are initialized to 0
          TransitionPointList *tpl;
          // bottom
          tpl = & (*mTData.begin()).second.mTrans;
          for (tliter=tpl->begin(); tliter!=tpl->end(); ++tliter)
            if ((*tliter).mTransition==EndAtEdge ||
                (*tliter).mTransition==Max ||
                (*tliter).mTransition==NoTransition)
              return false;
            else
              (*tliter).mBeforeAfter[0]=0;
          // top
          tpl = & (*(--mTData.end())).second.mTrans;
          for (tliter=tpl->begin(); tliter!=tpl->end(); ++tliter)
            if ((*tliter).mTransition==StartAtEdge ||
                (*tliter).mTransition==Min ||
                (*tliter).mTransition==NoTransition)
              return false;
            else
              (*tliter).mBeforeAfter[1]=0;
        }

        int i;
        // interesting cases are between top and bottom
        for (PanelIter curpanel=mPanels.begin(); curpanel!=mPanels.end(); 
             ++curpanel)
          {
            for (i=0; i<2; ++i)
              {
                int valencecount=(*curpanel).numlines;
                TDataIter tline = (*curpanel).beginend[i];
                bool cleanup=false;
                for (tliter = (*tline).second.mTrans.begin();
                     tliter != (*tline).second.mTrans.end(); ++tliter)
                  {
#ifndef NDEBUG
                    (*tliter).mBeforeAfter[1-i]= -1;
#endif
                    switch ((*tliter).mTransition)
                      {
                        case NoTransition:
                          (*tliter).mBeforeAfter[1-i]= 1;
                          break;

                        case StartAtEdge:
                          if (i)
                            (*tliter).mBeforeAfter[0]=0;
                          else
                            (*tliter).mBeforeAfter[1]=1;
                          break;

                        case EndAtEdge:
                          if (i)
                            (*tliter).mBeforeAfter[0]=1;
                          else
                            (*tliter).mBeforeAfter[1]=0;
                          break;

                        case Max:
                          if (i)
                            (*tliter).mBeforeAfter[0]=2;
                          else
                            (*tliter).mBeforeAfter[1]=0;
                          break;

                        case Min:
                          if (i)
                            (*tliter).mBeforeAfter[0]=0;
                          else
                            (*tliter).mBeforeAfter[1]=2;
                          break;

                        case Isolated:
                          (*tliter).mBeforeAfter[1-i]= 0;
                          break;

                        default:
                          assert((*tliter).mTransition==HorizontalAtEdge ||
                                 (*tliter).mTransition==Horizontal       ||
                                 (*tliter).mTransition==UnknownAtEdge    ||
                                 (*tliter).mTransition==Unknown);
                          assert((*curpanel).guidelines[i] != mTData.end());
                          if (i)
                            (*tliter).mBeforeAfter[0]=0;
                          else
                            (*tliter).mBeforeAfter[1]=0;
                          cleanup=true;
                          break;

                      }
                    valencecount -= (*tliter).mBeforeAfter[1-i];
                    assert((*tliter).mBeforeAfter[1-i]!= -1);
                  }
                assert(((*curpanel).guidelines[i] != mTData.end())==
                       cleanup);
                if (cleanup)
                  {
                    TDataIter guideline = (*curpanel).guidelines[i];
                    assert(guideline != mTData.end());
                    assert((*guideline).second.mTrans.size()<=3);
                    int whichintersection[3];
                    int intersectioncount[3];
                    int j=0,k;
                    ScalarType tdelta = ((*tline).first +
                                         (*tline).second.mLastT)*.5
                                        - (*guideline).first;
                    TransitionPointIter gliter;
                    for (gliter = (*guideline).second.mTrans.begin();
                         gliter != (*guideline).second.mTrans.end(); 
                         ++gliter, ++j)
                      {
                        assert(Matrix::Abs((*gliter).mNormal[0])>mEpsilonR);
                        ScalarType sintersection = (*gliter).mS -
                          tdelta * (*gliter).mNormal[1] /
                          (*gliter).mNormal[0];
                        (*curpanel).sintersections[i][j] = sintersection;
                        ScalarType smallestdist = 2;
                        for (tliter=(*tline).second.mTrans.begin(), k=0;
                             tliter != (*tline).second.mTrans.end(); 
                             ++tliter, ++k)
                          {
                            ScalarType dist = Matrix::Abs(sintersection-
                              (*tliter).mS);
                            if (dist<smallestdist)
                              {
                                smallestdist=dist;
                                whichintersection[j]=k;
                              }
                          }
                        assert(smallestdist<1.1);
                      }
                    for (k=0; k<(*tline).second.mTrans.size(); ++k)
                      intersectioncount[k]=0;
                    for (j=0; j<(*guideline).second.mTrans.size();++j)
                      {
                        assert(whichintersection[j]>=0);
                        assert(whichintersection[j]<
                                (*tline).second.mTrans.size());
                        assert(j==0 || whichintersection[j]>=
                                        whichintersection[j-1]);
                        ++intersectioncount[whichintersection[j]];
                      }
                    if ((*tline).first != (*tline).second.mLastT)
                      for (gliter = (*guideline).second.mTrans.begin(), j=0; 
                           j<(*guideline).second.mTrans.size(); ++gliter, ++j)
                        {
                          assert(gliter!=(*guideline).second.mTrans.end());
                          assert(whichintersection[j]>=0);
                          assert(whichintersection[j]<
                                  (*tline).second.mTrans.size());

                          tliter=(*tline).second.mTrans.begin();
                          std::advance(tliter, whichintersection[j]);
                          ScalarType tdelta = (*tliter).mT - (*guideline).first;
                          (*curpanel).sintersections[i][j] = 
                            (*gliter).mS - tdelta * (*gliter).mNormal[1] /
                            (*gliter).mNormal[0];
                        }
                    for (tliter = (*tline).second.mTrans.begin(), k=0;
                         tliter != (*tline).second.mTrans.end(); 
                         ++tliter, ++k)
                      {
                        if ((*tliter).mTransition==HorizontalAtEdge ||
                            (*tliter).mTransition==Horizontal       ||
                            (*tliter).mTransition==UnknownAtEdge    ||
                            (*tliter).mTransition==Unknown)
                          (*tliter).mBeforeAfter[1-i]=intersectioncount[k];
                        else if ((*tliter).mBeforeAfter[1-i]!=
                                  intersectioncount[k])
                          { // something went wrong!
                            // try a different orientation
                            return false;
                          }
                      }
                  }
                else if (valencecount!=0)
                  { // something went wrong!
                    // try a different orientation
                    return false;
                  }
              }
          }
        return true;
      }

        // build strands
    
    bool Cubic2dContourFinder 
    ::BuildStrands()
      {
        PanelIter curpanel=mPanels.begin();
        TDataIter tline=(*curpanel).beginend[0];
        //TDataIter guidelines[2];
        //guidelines[0]=mTData.end();
        //guidelines[1]=(*curpanel).guidelines[0];
        ScalarType *sintersections[2];
        sintersections[0]=0;
        sintersections[1]=(*curpanel).sintersections[0];
        bool last=false;
        int lines[2] = { 0, (*curpanel).numlines };
        while (true)
          { // sintersections[1]
            // tline               ^ increasing t
            // sintersections[0]   |
            int linesused[2] = { 0, 0 };     
            int turningpointnum=0;
            for (TransitionPointIter tliter=(*tline).second.mTrans.begin(); 
                 tliter!=(*tline).second.mTrans.end(); ++tliter)
              {
                mStrands.StarSize(linesused[1], 
                  (*tliter).mBeforeAfter[0], turningpointnum);
                int valence = (*tliter).mBeforeAfter[0] +
                              (*tliter).mBeforeAfter[1];
                assert(valence<=6);
                TransitionType transition=(*tliter).mTransition;
                int j,k;
                if (transition==UnknownAtEdge && valence>1)
                  { // Critical point so valence can be >2.
                    // However, if valence is 6, at least
                    // 2 are in adjacent triangle.
                    // Valence in this triangle could be odd, 
                    // since an odd number of edges could be 
                    // in the adjacent triangle.
                    assert(valence<=4);
                    // ******
                    // Could do something complicated, but that is 
                    // unlikely to be productive since it won't 
                    // necessarily be consistent with adjacent triangle.
                    // So I'm going to do something simple instead and 
                    // resolve the ambiguity when glueing the 
                    // triangles together
                    // ******
                    for (j=0; j<(*tliter).mBeforeAfter[0]; ++j)
                      mStrands.EndOneStrand(&(*tliter),
                                   curpanel, linesused[1]);
                    for (j=0; j<(*tliter).mBeforeAfter[1]; ++j)
                      mStrands.CreateOneStrand(&(*tliter), curpanel,
                                   linesused[1]+j, turningpointnum);
                  }
                else if (valence<=2)
                  {
#ifndef NDEBUG
                    if (transition==HorizontalAtEdge)
                      {
                        // probably can only be 0 if along side 0
                        // or in the top corner (side==5)
                        // In the top corner HorizontalAtEdge is converted
                        // to Isolated now.
                        assert(valence<=1);
                        assert(valence==1 || (*tliter).mSide==0);
                        //                  || (*tliter).mSide==5);
                      }
                    else if (transition==Horizontal)
                      {
                        // proably can only be < 2 if along side 0
                        assert(valence<=2);
                        assert(valence==2 || (*tliter).mSide==0);
                      }
#endif
                    if (transition==Unknown && (valence&1))
                      return false;
                    // connect lines according to
                    // (*tliter).mBeforeAfter[0..1]
                    // ...
                    switch (valence)
                      {
                        case 0:
                          // output isolated point
                          // possibly link to an edge or corner
                          mStrands.CreateOneStrand(&(*tliter), curpanel,
                                        linesused[1], turningpointnum);
                          mStrands.EndOneStrand(&(*tliter), 
                                         curpanel, linesused[1]);
                          if ((*tliter).mTransition==UnknownAtEdge)
                            (*tliter).mTransition=DeletePoint;
                          else
                            (*tliter).mTransition=Isolated;
                          break;
                        case 1:
                          // connect edge to a side
                          if ((*tliter).mBeforeAfter[0])
                            mStrands.EndOneStrand(&(*tliter),
                                         curpanel, linesused[1]);
                          else
                            mStrands.CreateOneStrand(&(*tliter), 
                                curpanel, linesused[1], turningpointnum);
                          break;
                        case 2:
                          // connect the two edges
                          if ((*tliter).mBeforeAfter[0]==2)
                            {
                              mStrands.JoinTwoStrands(&(*tliter),
                                         curpanel, linesused[1]);
                            }
                          else if ((*tliter).mBeforeAfter[1]==2)
                            {
                              mStrands.CreateTwoStrands(&(*tliter), 
                                curpanel, linesused[1], turningpointnum);
                            }
                          break;
                      }
                  }
                    // deal with guidelines, resolve ambiguity
                    // (*tliter).mBeforeAfter[0..1] set above
                    // also assign edges to endpoints
                else if (transition==Unknown)
                  { // critical point so valence can be >2
                    // must be even (verify!)
                    assert(valence%2==0);
                    // either valence == 4 or 6
                    // 1+3, 2+2, 3+1, or 3+3

                    assert(curpanel!=mPanels.begin());
                    assert(curpanel!=mPanels.end());
                    assert(sintersections[0]);
                    assert(sintersections[1]);

                    StarData  circle[6];
                    for (j=0; j<(*tliter).mBeforeAfter[0]; ++j)
                      circle[j].mS=sintersections[0]
                          [linesused[0]+j];
                    for (; j<valence; ++j)
                      circle[j].mS=sintersections[1]
                          [linesused[1]+valence-j-1];
                    for (j=0; j<valence; ++j)
                      circle[j].mNum = j;
                    std::sort(&circle[0], &circle[j]);
                    bool parity[2];
                    for (k=0; k<valence/2; ++k)
                      {
                        for (j=0; j<2; ++j)
                          {
                            int index= j ? valence-1-k : k;
                            parity[j]=((circle[index].mNum & 1) == 1);
                            if (circle[index].mNum >= 
                                (*tliter).mBeforeAfter[0])
                              parity[j] = ! parity[j];
                          }
                        if (parity[0]!=parity[1])
                          break;
                      }
                    // hopefully parity[0]!=parity[1] now
                    // would be best to somehow (?) use the location
                    // of (*tliter) as a tie-breaker
                    // another idea is to make the number of 
                    // left and right vertices equal by moving
                    // the root to the left or right of the
                    // median; this will change one of
                    // parity[0] and parity[1] so they will
                    // now be consistent
                    // if (parity[0]==parity[1])
                    //   DO THIS...


                    // if parity[0] then 0--1  2--3  [4--5]
                    // else 1--2  3--0  or  1--2  3--4  5--0
                    if (parity[0])
                      {
                        if ((*tliter).mBeforeAfter[0]>1)
                          mStrands.JoinTwoStrands(&(*tliter),
                                 curpanel, linesused[1]);
                        if ((*tliter).mBeforeAfter[1]>1)
                          mStrands.CreateTwoStrands(&(*tliter), curpanel, 
                                    linesused[1], turningpointnum);
                      }
                    else
                      {
                        if ((*tliter).mBeforeAfter[0]>2)
                          mStrands.JoinTwoStrands(&(*tliter),
                                 curpanel, linesused[1]+1);
                        if ((*tliter).mBeforeAfter[1]>2)
                          mStrands.CreateTwoStrands(&(*tliter), curpanel,
                                    linesused[1]+1, turningpointnum);
                      }
                  }
                linesused[0]+=(*tliter).mBeforeAfter[0];
                linesused[1]+=(*tliter).mBeforeAfter[1];
                ++turningpointnum;
              }
            assert(linesused[0]==lines[0]);
            assert(linesused[1]==lines[1]);

            if (curpanel==mPanels.end())
              {
                assert(mStrands.mFrontier.empty());
                break;
              }

            mStrands.EndSegment();

            //guidelines[0]=(*curpanel).guidelines[1];
            sintersections[0]=(*curpanel).sintersections[1];
            tline=(*curpanel).beginend[1];
            assert((*curpanel).numlines==lines[1]);
            lines[0]=(*curpanel).numlines;
            ++curpanel;
            if (curpanel==mPanels.end())
              {
                //guidelines[1]=mTData.end();
                sintersections[1]=0;
                lines[1]=0;
              }
            else
              {
                //guidelines[1]=(*curpanel).guidelines[0];
                sintersections[1]=(*curpanel).sintersections[0];
                lines[1]=(*curpanel).numlines;
              }
          }
        return true;
      }

    
    void Cubic2dContourFinder 
    ::SplineForStrands()
      {
        for (StrandCollection ::StrandIter curstrand =
                              mStrands.mStrands.begin(); 
              curstrand != mStrands.mStrands.end(); ++curstrand)
          {
            // In Pass 1, we iterate through each point in the
            // strand and figure out (as much as we can just examining
            // the one point) the position, tangent, and curvature
            // of the contour at those points
            assert((*curstrand).mTurningNum.size()==
                   (*curstrand).mStrandNum.size()+1);
            int panelsequence = 0;
            int curnum = (*curstrand).mTurningNum[panelsequence];

            PanelIter curpanel = (*curstrand).mPanelSpan[0];

            TDataIter curt;
            if (curpanel==mPanels.end())
              {
                curt = mTData.end();
                --curt;
              }
            else
              curt = (*curpanel).beginend[0];

            int nextnum;
            if ((*curstrand).mStrandNum.size()==0)
              nextnum = -1;
            else
              nextnum = (*curstrand).mStrandNum[panelsequence];

            TransitionPointIter tliter;
#ifndef NDEBUG
            tliter=(*curt).second.mTrans.begin();
            std::advance(tliter, curnum);
            assert(&(*tliter) == (*curstrand).mEndPts[0]);
#endif
            typedef std::list< Pass1Data > Pass1List;
            typedef Pass1List::iterator Pass1Iter;
            Pass1List pass1;
            
            while (true)
              {
                tliter=(*curt).second.mTrans.begin();
                std::advance(tliter, curnum);
                // cases: 
                //  mTransition == Unknown
                //      should skip this point, don't have curvature
                //  mTransition == UnknownAtEdge
                //      should use dummy tangent and curvature values
                //  mBeforeAfter[0] == 2 (MAX)
                //  mBeforeAfter[1] == 2 (MIN)
                //      tangent is horizontal, either left or right
                //      need to look at connected strand to resolve
                //  mTransition == Horizontal
                //  mTransition == HorizontalAtEdge
                //      need to compare with other point to resolve
                //      whether tangent is left or right
                //  Otherwise
                //      tangent direction easy from sign of gradient
                //  may need to flip sign of curvature!
                if ((*tliter).mTransition==Unknown)
                  {
                    // mark some point only needing a join
                    // (rather than the full OptimizeSplinePath treatment)
                    if (! pass1.empty())
                      pass1.back().mTransition=Unknown;
                    // note that two interesting points can't be adjacent
                    // in the list of Transition Points
                    // don't worry about an Unknown at the beginning of the pass1
                    // will be dealt with when strands are connected
                  }
                else if ((*tliter).mTransition==DeletePoint)
                  ;
                else
                  {
                    pass1.push_back();
                    Pass1Data *datum= & pass1.back();
                    datum->mRootNum=nextnum;
                    if (curpanel==(*curstrand).mPanelSpan[1])
                      datum->mTotalRoots=0;
                    else
                      datum->mTotalRoots=(*curpanel).numlines;
                    assert(datum->mRootNum < datum->mTotalRoots);
                    datum->mPoint.mST(0,0) = (*tliter).mS;
                    datum->mPoint.mST(0,1) = (*tliter).mT;

                    if ((*tliter).mTransition==UnknownAtEdge)
                      {
                        datum->mTransition=UnknownAtEdge;
                        // don't need mCurvature and mTangent
                        // will construct spline segment without them below
                      }
                    else if ((*tliter).mBeforeAfter[0]==2 || 
                             (*tliter).mBeforeAfter[1]==2)
                      {
                        datum->mTransition=NoTransition;
                        // need to figure out if horizontal tangent
                        // if it should point left or right
                        datum->mPoint.mTangent(0,1) = 0;
                        // either the beginning or ending of curstrand
                        // see if can figure out sequence number
                        // of connected strand
                        if ((*tliter).mBeforeAfter[0]==2)
                          { // maximum
                            assert((*tliter).mBeforeAfter[1]==0);
                            Strand *otherstrand = & mStrands.mStrands[
                              (*curstrand).mConnectedTo[1] ];
                            if (otherstrand->mStrandNum.back() <
                                (*curstrand).mStrandNum.back())
                              datum->mPoint.mTangent(0,0) = - 1;
                            else
                              datum->mPoint.mTangent(0,0) = + 1;
                          }
                        else
                          { // minimum
                            assert((*tliter).mBeforeAfter[0]==0);
                            assert((*tliter).mBeforeAfter[1]==2);
                            assert((*curstrand).mConnectedTo[0]>=0);
                            assert((*curstrand).mConnectedTo[0]<
                                   mStrands.mStrands.size());
                            Strand *otherstrand = & mStrands.mStrands[
                              (*curstrand).mConnectedTo[0] ];
                            if (otherstrand->mStrandNum.front() <
                                (*curstrand).mStrandNum.front())
                              datum->mPoint.mTangent(0,0) = + 1;
                            else
                              datum->mPoint.mTangent(0,0) = - 1;
                          }
                        if (datum->mPoint.mTangent(0,0)*
                            (*tliter).mNormal[1] > 0)
                          datum->mPoint.mCurvature=  (*tliter).mCurvature;
                        else
                          datum->mPoint.mCurvature= -(*tliter).mCurvature;
                      }
                    else if ((*tliter).mTransition==Horizontal ||
                             (*tliter).mTransition==HorizontalAtEdge)
                      {
                        datum->mTransition=Horizontal;
                        // curvature is 0, otherwise would have
                        // classified as a max or a min
                        datum->mPoint.mCurvature= 0;
                        // will set datum->mPoint.mTangent later
                        // know that it will be either unit left or right
                        // since horizontal
                        datum->mPoint.mTangent(0,1) = 0;
                      }
                    else
                      {
                        datum->mTransition=NoTransition;
                        TransitionToSplinePoint(*tliter, datum->mPoint);
                      }
                  }

                // advance iterators...
                if (nextnum==-1)
                  break;
                curnum=nextnum;
                ++curt;
                if (curt==(*curpanel).beginend[1])
                  {
                    ++curpanel;
                    ++panelsequence;

                    assert((*curstrand).mTurningNum.size()>panelsequence);
                    curnum=(*curstrand).mTurningNum[panelsequence];
                    if ((*curstrand).mStrandNum.size()>panelsequence)
                      nextnum = (*curstrand).mStrandNum[panelsequence];
                    else
                      {
                        assert(curpanel==(*curstrand).mPanelSpan[1]);
                        nextnum = -1;
                      }
                  }

                assert(curnum!=-1);
              }
            // In Pass 2, we deal with two consecutive points at
            // a time, resolve anything that could not be resolved
            // looking at just one point, and output a spline segment
            // connecting those two points
            Pass1Iter segment[2];
            segment[0]=pass1.begin();
            assert(segment[0]!=pass1.end());
            segment[1]=segment[0];
            ++segment[1];
            if (segment[1]==pass1.end())
              { // isolated point
                (*curstrand).mControlPolygon.push_back((*segment[0]).mPoint.mST);
                (*curstrand).mControlPolygon.push_back((*segment[0]).mPoint.mST);
                (*curstrand).mControlPolygon.push_back((*segment[0]).mPoint.mST);
              }
            
            for (; segment[1]!=pass1.end(); 
                  segment[0]=segment[1], ++segment[1])
              {
                bool done=false;
                for (int signum=0; signum<2; ++signum)
                  if ((*segment[signum]).mTransition==UnknownAtEdge)
                    { // a quadratic curve is completely determined by
                      // the location of the two endpoints, and the tangent and
                      // curvature at one of the endpoints
                      STLocation c1, c2;
                      UnknownToKnown((*segment[signum]).mPoint, 
                                     (*segment[1-signum]).mPoint, c1, c2);
                      (*curstrand).mControlPolygon.push_back((*segment[0]).mPoint.mST);
                      if (signum)
                        {
                          (*curstrand).mControlPolygon.push_back(c2);
                          (*curstrand).mControlPolygon.push_back(c1);
                        }
                      else
                        {
                          (*curstrand).mControlPolygon.push_back(c1);
                          (*curstrand).mControlPolygon.push_back(c2);
                        }
                      done=true;
                    }
                  else if ((*segment[signum]).mTransition==Horizontal)
                    {
                      // Unfortunately, not a totally reliable test.
                      // Luckily, the use of guidelines helps this
                      // test a lot
                      if ((*segment[1]).mPoint.mST(0,0) <
                          (*segment[0]).mPoint.mST(0,0))
                        (*segment[signum]).mPoint.mTangent(0,0) = -1;
                      else
                        (*segment[signum]).mPoint.mTangent(0,0) = 1;
                    }
                if (! done)
                  {
                    const SplinePoint  *pts[2];
                    pts[0]=& (*segment[0]).mPoint;
                    pts[1]=& (*segment[1]).mPoint;
                    if ((*segment[0]).mTransition==Unknown)
                      { // first point is Unknown implies
                        // this segment is only a join
                        STLocation controlpts[2];
                        SimpleJoin(pts, controlpts);
                        (*curstrand).mControlPolygon.push_back(pts[0]->mST);
                        (*curstrand).mControlPolygon.push_back(controlpts[0]);
                        (*curstrand).mControlPolygon.push_back(controlpts[1]);
                      }
                    else
                      {
                        OptimizeSplinePath((*segment[0]).mRootNum, 
                            (*segment[0]).mTotalRoots, pts,
                            (*curstrand).mControlPolygon);
                      }
                  }
              }
            // complete the curve
            (*curstrand).mControlPolygon.push_back((*segment[0]).mPoint.mST);
          }
      }

            // stitch together connected strands
            // also, rotate side numbers at beginning and
            // end of each contour
    
    void Cubic2dContourFinder 
    ::StitchConnectedStrands()
      {
        std::hash_set<int> strandsleft;
        int strandnum;
        for (strandnum=0; strandnum<mStrands.mStrands.size(); ++strandnum)
          strandsleft.insert(strandnum);

        while (! strandsleft.empty())
          {
            // choose one strand from set of strands left to process
            strandnum = * strandsleft.begin();
            //strandsleft.erase(strandsleft.begin());
            int firststrand = strandnum;
            int reversed=0;
            int nextstrand = mStrands.mStrands[firststrand].
                                        mConnectedTo[reversed];
            // DEBUG
            int prevstrand = mStrands.mStrands[firststrand].
                                        mConnectedTo[1-reversed];
            if (prevstrand>-1)
              {
                int prevstrand0 = mStrands.mStrands[prevstrand].
                                        mConnectedTo[0];
                int prevstrand1 = mStrands.mStrands[prevstrand].
                                        mConnectedTo[1];
              }
            while (nextstrand != -1 && nextstrand != strandnum)
              {
                firststrand=nextstrand;
                reversed=1-reversed;
                nextstrand = mStrands.mStrands[firststrand].
                                        mConnectedTo[reversed];
              }
            (*mResults).push_back();
            if (nextstrand==strandnum)
              { // closed loop
                (*mResults).back().mEndingSides[0]=-1;
                (*mResults).back().mEndingSides[1]=-1;
              }
            else
              { // topologically linear
                (*mResults).back().mEndingSides[0] = RotateSide(mGoodSide,
                        mStrands.mStrands[firststrand].mEndPts[reversed]->mSide);
                assert((*mResults).back().mEndingSides[reversed] != -1);
              }
            nextstrand=firststrand;
            Strand *curstrand;
            do
              {
                strandnum=nextstrand;
                curstrand=& mStrands.mStrands[strandnum];
                strandsleft.erase(strandnum);

                if (curstrand->mEndPts[reversed]->mTransition == Unknown)
                  { // need to add a join between two segments at a max or min
                    // tline of first and last point: 
                    // curstrand->mPanelSpan[reversed].guidelines[reversed]
                    // indices are neighbor->mStrandNum.(
                    //      reversed ? back() : front())
                    // and curstrand->mStrandNum.(
                    //      reversed ? back() : front())

                    // reversed == 1
                    // neighbor /~~~\ curstrand

                    // reversed == 0
                    // neighbor \___/ curstrand
                    int prevstrand=curstrand->mConnectedTo[reversed];
                    assert(prevstrand!=-1);
                    Strand *neighbor=& mStrands.mStrands[prevstrand];
                    PanelIter curpanel=curstrand->mPanelSpan[reversed];
                    if (reversed)
                      --curpanel;
#ifndef NDEBUG
                    PanelIter nbrpanel=neighbor->mPanelSpan[reversed];
                    if (reversed)
                      --nbrpanel;
                    assert((*curpanel).guidelines[reversed]==(*nbrpanel). 
                            guidelines[reversed]);
#endif
                    assert((*curpanel).guidelines[reversed]!=mTData.end());
                    TDataIter tline=(*curpanel).guidelines[reversed];
                    const TransitionPointType *endpts[2];
                    int strandnums[2];
                    bool increasingt[2];
                    if (reversed)
                      {
                        strandnums[0]=neighbor->mStrandNum.back();
                        strandnums[1]=curstrand->mStrandNum.back();
                        increasingt[0]=true;
                        increasingt[1]=false;
                      }
                    else
                      {
                        strandnums[0]=neighbor->mStrandNum.front();
                        strandnums[1]=curstrand->mStrandNum.front();
                        increasingt[0]=false;
                        increasingt[1]=true;
                      }
                    for (int i=0; i<2; ++i)
                      {
                        TransitionPointIter tliter=(*tline).second.mTrans.begin();
                        std::advance(tliter,  strandnums[i]);
                        endpts[i]=&(*tliter);
                      }

                    // call MinMaxJoin
                    STLocation controlpts[2];
                    MinMaxJoin(endpts, (*tline).first, increasingt, 
                                controlpts);

                    // output first/last element of control polygon of
                    //  neighbor
                    OutputVertex(*mResults, 
                         reversed ? neighbor->mControlPolygon.back() :
                                    neighbor->mControlPolygon.front());
                    // then controlpts[0], controlpts[1]
                    OutputVertex(*mResults, controlpts[0]);
                    OutputVertex(*mResults, controlpts[1]);
                  }
                if (reversed)
                  for (int i=curstrand->mControlPolygon.size()-1; i>0; --i)
                    OutputVertex(*mResults, curstrand->mControlPolygon[i]);
                else
                  for (int i=0; i<curstrand->mControlPolygon.size()-1; ++i)
                    OutputVertex(*mResults, curstrand->mControlPolygon[i]);
                reversed=1-reversed;
                nextstrand=curstrand->mConnectedTo[reversed];
              } while (nextstrand!=-1 && nextstrand!=firststrand);
            //if (nextstrand==firststrand && 
            //    curstrand->mEndPts[reversed]->mTransition == Unknown)
            // maybe need to add a join to close a loop
            OutputVertex(*mResults, reversed ?
                                    curstrand->mControlPolygon.back() :
                                    curstrand->mControlPolygon.front());
            if (nextstrand==-1)
              {
                assert((*mResults).back().mEndingSides[0]!=-1);
                (*mResults).back().mEndingSides[1]=RotateSide(mGoodSide,
                        curstrand->mEndPts[reversed]->mSide);
              }
          }
      }

    
    void Cubic2dContourFinder 
    ::OutputZeroSides()
      {
        // Tricky since mZeroSides has not been rotated
        // by mGoodSide, but mRootsOnSides and mCubic have
        for (int side=0; side<3; ++side)
          if (mZeroSides[side])
            {
              int rotatedside=RotateSide((3-mGoodSide)%3, side);
              int rotatedp1=(rotatedside+1)%3;
              STLocation st[3];
              st[0](0,0)=0;    st[0](0,1)=0;
              st[1](0,0)=1;    st[1](0,1)=0;
              st[2](0,0)=0;    st[2](0,1)=1;

              int sidep4=side+4;
              if (sidep4>5) 
                sidep4 -= 3;
              if (mRootsOnSides[rotatedside].empty())
                {
                  OutputStraightLineSegment(*mResults,
                                            st[rotatedside], st[rotatedp1],
                                            side+3,
                                            sidep4);
                }
              else
                {
                  std::list<ContourType> outer;
                  STLocation start=st[rotatedside], finish=st[rotatedp1];
                  STLocation prev=start, cur;
                  int prevside=side+3;
                  for (RootIter iter=mRootsOnSides[rotatedside].begin();
                       iter != mRootsOnSides[rotatedside].end(); ++iter)
                    if (*iter>mEpsilon && *iter<1-mEpsilon)
                      {
                        // cur = start + (*iter)*(finish-start);
                        cur = start;
                        Matrix::AddScaledDiff(cur, *iter, finish, start);

                        OutputStraightLineSegment(*mResults, 
                                                  prev, cur,
                                                  prevside, 
                                                  side);
                        prev=cur;
                        prevside=side;
                      }
                  OutputStraightLineSegment(*mResults, 
                                            prev, finish,
                                            prevside, 
                                            sidep4);
                }
            }
      }

    bool Cubic2dContourFinder::TryAnOrientation()
      {
        ComputeGradients();

        // sort intersections with sides and roots of f=0 and 
        // partial f/partial s =0 on each side in order of 
        //    increasing "t"
        InsertEdgePoints();
        InsertInteriorPoints();
        FindRootsAtSameT();
        RemoveDeletedPoints();

        // check to see if we have anything to contour
        if (! mTData.empty())
          {
            RemoveRedundantInterior();

            InitializePanels();
            if (mPanels.empty())
              {
                if (!NoPanelsCase())
                  return false;
              }
            else
              {
                ComputeGuideLines();
                if (!SetBeforeAfter())
                  return false;

                // for each panel connect vertices on starting edge with
                // vertices on ending edge, with topology from above
                if (!BuildStrands())
                  return false;
                SplineForStrands();
                StitchConnectedStrands();
              }
          }
        return true;
      }

    // Implements Grandine1997 modified for triangular cubic patches

    Cubic2dContourFinder 
    ::Cubic2dContourFinder(const CubicType &cubic, ScalarType epsilon, 
                      ScalarType epsilonR, std::list<ContourType> & results)
      {
        mCubic=cubic;
        mEpsilon=epsilon;
        mEpsilonR=epsilonR;
        mResults=&results;

        while (1)
          {
            FactorOutZeroSides();
            FindRootsOnEachSide();
            mGoodSide=0;
            int tryagain;
            while (!(tryagain=DetermineGoodSide()))
              {
                if (TryAnOrientation())
                  {
                    OutputZeroSides();
                    return;
                  }
                (*mResults).clear();
                mTData.clear();
                mPanels.clear();
                IncrementGoodSide();
              }
            if (tryagain==2)
              {
                mCubic=cubic;
                mEpsilon/=8;
                mEpsilonR/=8;
                mRootsOnSides[0].clear();
                mRootsOnSides[1].clear();
                mRootsOnSides[2].clear();
              }
            else 
              return;
          }
      }

    
    void FindContour(const TriangularCubic 
                     &cubic, ScalarType epsilon, ScalarType epsilonR,
                     std::list<Contour> & results)
      {
        {
          bool pos;
          int i;
          for (i=0; i<10; ++i)
            if (cubic.mY[i]>epsilonR)
              {
                pos=true;
                break;
              }
            else if (cubic.mY[i]<-epsilonR)
              {
                pos=false;
                break;
              }
          if (i==10)
            // All 0!
            return;
          if (i==0)
            {
              if (pos)
                {
                  for (++i; i<10; ++i)
                    if (cubic.mY[i]<epsilonR)
                      break;
                }
              else
                {
                  for (++i; i<10; ++i)
                    if (cubic.mY[i]>-epsilonR)
                      break;
                }
              if (i==10)
                { // all positive or all negative
                  return;
                }
            }
        }

        Cubic2dContourFinder  
          (cubic, epsilon, epsilonR, results);
      }
  }
