/* int1d.h */
#ifndef __INTERPOLANT_CUBIC_1D_H
#define __INTERPOLANT_CUBIC_1D_H

#include "int1dbs.h"
#include "func1d.h"
#include "cubic.h"
#include "root1d.h"

class InterpolantCubic1D
: public InterpolantCubic1DBase<VertexInfo<DomainType1d, Extra_1D_Cubic> >
  {
public:
    typedef InterpolantCubic1D thistype;
    typedef FunctionCubic1D FunctionType;
    FunctionType* mFunction;
    ScalarType mConservativeError;

    void EvaluateElement(MeshElement_VirtualBase<VertexType> *element)
      {
      }

    void Evaluate(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::ValueEvaluated));
        assert(! (vert.mEvaluated & VertexType::InterpolationDataEvaluated));
        vert.mFunctionValue=mFunction->EvaluateValue(vert.mPos(0,0));
        vert.mDerivative=mFunction->EvaluateDerivative(vert.mPos(0,0));
        vert.mEvaluated |= VertexType::ValueEvaluated |
                           VertexType::InterpolationDataEvaluated;
      }
    void EvaluateCorner(VertexType &vert)
      {
        if (! (vert.mEvaluated & VertexType::ValueEvaluated))
          Evaluate(vert);
      }

    void EvaluateCenter(VertexType &vert)
      {
        Evaluate(vert);
      }

    ElementAction AnalyzeElement(MeshElement_VirtualBase<VertexType> *element, 
                                 ScalarType radius, VertexType *center, 
                                 ScalarType epsilon)
      {
        assert(radius>0);
        assert(epsilon>=0);

        VertexType *v[2];
        for (int i=0; i<2; ++i)
          {
            v[i]=& element->Corner(i);
            assert(v[i]->mEvaluated == VertexType::BothEvaluated);
          }

        ScalarType error;
          {
            // value at center of element = average of values at ends
            // plus radius*(f'(0)-f'(2r))/16
            ScalarType val = 
              (v[0]->mFunctionValue+v[1]->mFunctionValue)*.5
              +(v[0]->mDerivative-v[1]->mDerivative)*radius*(1./4);

            // derivative at center of element = (f(2r)-f(0))/(2*radius)
            // plus (f'(2r)-f'(0))/4
            ScalarType deriv = 
              (v[1]->mFunctionValue-v[0]->mFunctionValue)*.75/radius
              -(v[1]->mDerivative+v[0]->mDerivative)*(1./4);

            assert(center->mEvaluated == VertexType::BothEvaluated);
            val = Matrix::Abs(val - center->mFunctionValue);
            deriv = Matrix::Abs(deriv - center->mDerivative);
            error=mConservativeError*ErrorOfCubic(val, 2*deriv, radius);
          }

        assert(error>=0);

        if (Root1D::SignOfProduct(v[0]->mFunctionValue, 
                                  v[1]->mFunctionValue)==Root1D::Positive)
          {
            if (v[0]->mFunctionValue>0)
              {
                if (v[0]->mFunctionValue+v[0]->mDerivative*radius*(2./3)-error
                                                            > 0 &&
                    v[1]->mFunctionValue-v[1]->mDerivative*radius*(2./3)-error
                                                            > 0)
                  return DiscardElement;
              }
            else
              {
                if (v[0]->mFunctionValue+v[0]->mDerivative*radius*(2./3)+error
                                                            < 0 &&
                    v[1]->mFunctionValue-v[1]->mDerivative*radius*(2./3)+error
                                                            < 0)
                  return DiscardElement;
              }
          }
        if (StoppingCriteriaHelper(error, epsilon, element))
          return ContourElement;
        return RefineElement;
      }
  };

#endif
