#ifndef __INTERPOLANT_CUBIC_1D_BASE_H
#define __INTERPOLANT_CUBIC_1D_BASE_H

#include "template.h"
#include "smallmat.h"
#include "vert.h"
//#include "matrix.h"
//#include "meshelt.h"
//#include "root1d.h"
//#include <assert.h>
//#include <math.h>
#include "traits1d.h"

struct Extra_1D_Cubic
  {
    ScalarType mDerivative;
  };

template<class Vertex_T>
class InterpolantCubic1DBase
  {
public:
    typedef Vertex_T VertexType;
    typedef InterpolantTraits1D<VertexType> TraitsType;
    typedef TraitsType::ContourType ContourType;

    ScalarType mEpsilon;
    ScalarType mEpsilonR;

    virtual void EvaluateElement(MeshElement_VirtualBase<VertexType> *element)=0;

    void InterpolateContour(std::list<ContourType> &results, 
                            MeshElementList_VirtualBase<VertexType> *elt_lst)
      {
        MeshElement_VirtualBase<VertexType> *element;
        while ((element=elt_lst->NextElement())!=0)
          {
            assert(element->NumCorners()==2);
            VertexType *v[2];

            EvaluateElement(element);
            for (int i=0; i<2; ++i)
              {
                v[i]=& element->Corner(i);
                assert(v[i]->mEvaluated == VertexType::BothEvaluated);
              }
            // Find all roots in interval with endpoints:
            // v[0]->mPos, v[0]->mFunctionValue, v[0]->mDerivative, 
            // v[1]->mPos, v[1]->mFunctionValue, v[1]->mDerivative
            // assuming cubic interpolation
            // call results.push_back with each
#if 0
            if (v[0]->mFunctionValue==0)
              Root1D::AddResult(results, v[0]->mPos(0,0));
            if (v[1]->mFunctionValue==0)
              Root1D::AddResult(results, v[1]->mPos(0,0));

            Root1D::Solve_Bisection(results, v[0]->mPos(0,0), v[0]->mFunctionValue, v[0]->mDerivative,
                                    v[1]->mPos(0,0)-v[0]->mPos(0,0), 
                                    v[1]->mFunctionValue, v[1]->mDerivative, mEpsilon);
#else
            Root1D::Hermite  
                        h(v[0]->mPos(0,0),      v[1]->mPos(0,0)-v[0]->mPos(0,0),
                          v[0]->mFunctionValue, v[1]->mFunctionValue,
                          v[0]->mDerivative,    v[1]->mDerivative);
            Root1D::BezierInterval  b;
            b.SetFromHermite(h);
#if 0
            Root1D::Solve_Bisection(results, b, mEpsilon);
#elif 0
            Root1D::Solve_QuadraticBox(results, b, mEpsilon, mEpsilonR);
#else
            Root1D::Solve_N_Box(results, b, mEpsilon, mEpsilonR);
#endif
#endif
          }
      }

protected:
    bool StoppingCriteriaHelper(ScalarType maxerr, ScalarType epsilon, 
                                MeshElement_VirtualBase<VertexType> *element)
      {
        assert(maxerr>=0);
        assert(epsilon>=0);
        assert(element);
        assert(element->NumCorners()==2);

        VertexType *v[2];
        for (int i=0; i<2; ++i)
          {
            v[i]=& element->Corner(i);
            assert(v[i]->mEvaluated == VertexType::BothEvaluated);
          }

        // Compute minimum derivative
        ScalarType xd = v[1]->mPos(0,0)-v[0]->mPos(0,0);
        ScalarType d0 = v[0]->mDerivative;
        ScalarType d2 = v[1]->mDerivative;
        ScalarType d1 = (v[1]->mFunctionValue-v[0]->mFunctionValue)*3/xd-d0-d2;
        ScalarType a0 = (d1-d0); // *2/xd
        ScalarType a1 = (d2-d1); // *2/xd
          // a0, a1 are proportional to initial and final accellerations
          // sufficient unless minderiv==0 below
        ScalarType minderiv;
        if ((d0>0) != (d2>0))
          {
            minderiv=0;
          }
        else
          {
            minderiv = Matrix::Min(Matrix::Abs(d0), Matrix::Abs(d2));
            if ((a0>0) != (a1>0))
              { // derivative has an extremum in the interval
                ScalarType t = a0/(a0-a1);
                ScalarType dtemp0 = d0 + (d1-d0)*t;
                ScalarType dtemp1 = d1 + (d2-d1)*t;
                ScalarType d = Matrix::Abs(dtemp0 + (dtemp1-dtemp0)*t);
                if (d<minderiv)
                  minderiv=d;
              }
          }

        // easy stopping criteria is 2*radius<=epsilon, 
        // but that is built into the contouring algorithm
        // i.e. equivalent to always returning false
        
        // better stopping criteria is to let md = min|f'(x)| over |x-center|<=radius
        // if md == 0 then use the second derivative stopping criteria
        // else let fe be an estimate of the approximation error
        // if epsilon > fe/md return true
        
        if (minderiv==0)
          { // could just return false, but we can actually do better
            if ((a0>0) != (a1>0) || a0==0 || a1==0)
              return false;

            ScalarType factor = 2/xd;
            a0 *= factor;
            a1 *= factor;

            ScalarType minaccel = Matrix::Min(Matrix::Abs(a0), Matrix::Abs(a1));
            assert(minaccel>0);
            return sqrt(maxerr/minaccel) < epsilon;
          }
        assert(minderiv>0);
        return epsilon*minderiv > maxerr;
      }
  };

#endif
