#ifndef __INTERPOLANT_LINEAR_1D_H
#define __INTERPOLANT_LINEAR_1D_H

#include "template.h"
#include "vert.h"
#include "traits1d.h"
#include "func1d.h"
#include "meshelt.h"
#include <list>

struct Extra_1D_Linear
  {
  };

class InterpolantLinear1D
  {
public:
    typedef VertexInfo<DomainType1d, Extra_1D_Linear> VertexType;
    typedef InterpolantTraits1D<VertexType> TraitsType;
    typedef TraitsType::ContourType ContourType;

    typedef FunctionLinear1D FunctionType;
    FunctionType *mFunction;

    ScalarType mConservativeError;
    ScalarType mThirdOrderError;
    ScalarType mEpsilon;
    ScalarType mEpsilonR;

    void Evaluate(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::ValueEvaluated));
        vert.mFunctionValue=mFunction->EvaluateValue(vert.mPos(0,0));
        vert.mEvaluated |= VertexType::ValueEvaluated |
                           VertexType::InterpolationDataEvaluated;
      }

    void EvaluateCorner(VertexType &vert)
      {
        if (! (vert.mEvaluated & VertexType::ValueEvaluated))
          Evaluate(vert);
      }

    void EvaluateCenter(VertexType &vert)
      {
        Evaluate(vert);
      }

    ElementAction AnalyzeElement(MeshElement_VirtualBase<VertexType> *element, 
                                 ScalarType radius, VertexType *center, 
                                 ScalarType epsilon);

    void InterpolateContour(std::list<ContourType> &results, 
                            MeshElementList_VirtualBase<VertexType> *elt_lst);
  };

#endif
