#include "int2d.h"
#include "cubic.h"

ElementAction Interpolant2D 
::AnalyzeCubicElement(MeshElement_VirtualBase<VertexType> *element, 
                 ScalarType radius, VertexType *center, 
                 ScalarType epsilon, ScalarType epsilonR, 
                 const CubicErrorData &err)
  {
    VertexType *v[4];
    int numcorners=element->NumCorners();
    int i;
    const ScalarType sqrt2=1.414214; // sqrt(2)

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
      }

    ScalarType error;
    if (numcorners==3)
      {
        DomainType edge, edgeunit, edgenormal;
        Matrix::Subtract(edge, v[2]->mPos, v[0]->mPos);
        ScalarType edgelen=2*radius;
        assert(Matrix::Abs(Matrix::TwoNorm(edge)-edgelen)<0.00001);
        Matrix::EqScale(edgeunit, 1/edgelen, edge);
        edgenormal(0,0)= -edgeunit(0,1);
        edgenormal(0,1)= edgeunit(0,0);
        ScalarType g[3][2];
        g[0][0] = Matrix::Dot(edgeunit, v[0]->mGradient);
        g[0][1] = Matrix::Dot(edgenormal, v[0]->mGradient);
        g[1][0] = Matrix::Dot(edgeunit, center->mGradient);
        g[1][1] = Matrix::Dot(edgenormal, center->mGradient);
        g[2][0] = Matrix::Dot(edgeunit, v[2]->mGradient);
        g[2][1] = Matrix::Dot(edgenormal, v[2]->mGradient);
        ScalarType val, d1, d2;
        val = (v[0]->mFunctionValue+v[2]->mFunctionValue)*.5
                  +(g[0][0]-g[2][0])*radius*(1./4);
        val = Matrix::Abs(center->mFunctionValue-val);
        d1 = (v[2]->mFunctionValue-v[0]->mFunctionValue)*.75/radius
                  -(g[2][0]+g[0][0])*(1./4);
        d2 = Matrix::Abs(g[1][1] - (g[0][1]+g[2][1])*.5);

        assert(val>=0);
        assert(d2>=0);
        error = val + (Matrix::Abs(d1)+d2)*(0.104757*sqrt2)*radius;
        // this is intended to be an over estimate
        //ScalarType dn = sqrt(d1*d1+d2*d2);
        //error=mConservativeError*ErrorOfCubic(val, 2*dn, radius*sqrt2);
        ScalarType radius2=radius*radius;
        error += radius2*radius2*radius*err.mFifthOrderError;
        error *= err.mConservativeError;
        return StoppingCriteriaHelperTriangle
          (error, epsilon, err.mOptimism, element);
      }
    else
      {
        assert(numcorners==4);
        Root2D::TriangularCubic cubics[4];
        Interp2D::SibsonSplitSquare(
          reinterpret_cast<Vertex_2D_Cubic_Base **>(v), cubics, epsilonR);
        //cubics[*].mY[0] should be compared to center->mFunctionValue
        ScalarType val=Matrix::Abs(cubics[0].mY[0] - 
                                    center->mFunctionValue);
        //cubics[*].mY[1,4] should be compared to center->mFunctionValue
        // + Dot(center->mGradient, delta x)
        DomainType edge1, edge2;
        Matrix::Subtract(edge1, cubics[0].mP[1], 
                                cubics[0].mP[0]);
        Matrix::Subtract(edge2, cubics[0].mP[2],
                                cubics[0].mP[0]);
        ScalarType d1, d2;
        d1 = center->mFunctionValue + Matrix::Dot(edge1, center->mGradient)/3;
        d1 = Matrix::Abs(d1 - cubics[0].mY[1])*3/radius;
        d2 = center->mFunctionValue + Matrix::Dot(edge2, center->mGradient)/3;
        d2 = Matrix::Abs(d2 - cubics[0].mY[4])*3/radius;
        assert(val>=0);
        assert(d1>=0);
        assert(d2>=0);
        error = val + (d1+d2)*(0.157784*sqrt2)*radius;
        //ScalarType dn = sqrt(d1*d1+d2*d2);
        //error=mConservativeError*ErrorOfCubic(val, 2*dn, radius*sqrt2);
        ScalarType radius2=radius*radius;
        error += radius2*radius2*radius*err.mFifthOrderError;
        error *= err.mConservativeError;
        return StoppingCriteriaHelperSquare
          (error, epsilon, err.mOptimism, element, cubics);
      }
  }

#if 0
bool Interpolant2D
::Cull(MeshElement_VirtualBase<VertexType> *element, 
                            ScalarType radius, ScalarType error)
  {
    // this is already in the stopping criteria
    VertexType *v[4];
    int numcorners=element->NumCorners();
    int i;

    bool allpos=true;
    bool allneg=true;
    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
        allpos = allpos && (v[i]->mFunctionValue>0);
        allneg = allneg && (v[i]->mFunctionValue<0);
      }
    static int tocheck[4] = { 2, 5, 6, 8 };
    if (allpos || allneg)
      if (numcorners==3)
        {
          Root2D::TriangularCubic cubics[3];
          Interp2D::CloughTocher(v, cubics, mEpsilonR);
          if (allpos)
            {
              for (i=0; i<3; ++i)
                for (int j=0; j<4; ++j)
                  if (cubics[i].mY[tocheck[j]]-error <= 0)
                    return false;
              return true;
            }
          else
            {
              assert(allneg);
              for (i=0; i<3; ++i)
                for (int j=0; j<4; ++j)
                  if (cubics[i].mY[tocheck[j]]+error >= 0)
                    return false;
              return true;
            }
        }
      else
        {
          assert(numcorners==4);
          assert(element==mLastSquareElement);
          /*
            *---*---*---*
            |\         /|
            | *   *   * |
            |  \     /  |
            *   *   *   *
            |    \ /    |
            | *   *   * |
            |    / \    |
            *   *   *   *
            |  /     \  |
            | 2   5   * |
            |/         \|
            *---6---8---*
           */
          if (allpos)
            {
              for (i=0; i<4; ++i)
                for (int j=0; j<4; ++j)
                  if (cubics[i].mY[tocheck[j]]-error <= 0)
                    return false;
              return true;
            }
          else
            {
              assert(allneg);
              for (i=0; i<4; ++i)
                for (int j=0; j<4; ++j)
                  if (cubics[i].mY[tocheck[j]]+error >= 0)
                    return false;
              return true;
            }
        }
    return false;
  }
#endif
