/* int2d.h */
#ifndef __INTERPOLANT_2D_H
#define __INTERPOLANT_2D_H

#include "int2dbs.h"
#include "func2d.h"

class Interpolant2D 
: public Interpolant2DBase<VertexInfo<DomainType2d, Extra_2D_Cubic>, 
                           InterpStrategy>
  {
public:
    typedef FunctionCubic2D FunctionType;
    struct CubicErrorData
      {
        ScalarType    mOptimism;
        ScalarType    mConservativeError;
        ScalarType    mFifthOrderError;
        CubicErrorData(ScalarType o, ScalarType c, ScalarType f)
          : mOptimism(o), mConservativeError(c), mFifthOrderError(f)
          { }
      } mErr;
    FunctionType *mFunction;

    Interpolant2D() : mErr(.3, 2, 1) // reasonable defaults
      { }

    void Evaluate(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::ValueEvaluated));
        assert(! (vert.mEvaluated & VertexType::InterpolationDataEvaluated));
        mFunction->ValueAndGrad(vert.mPos(0,0), vert.mPos(0,1), 
                                vert.mFunctionValue, vert.mGradient(0,0),
                                vert.mGradient(0,1));
        vert.mEvaluated |= VertexType::ValueEvaluated |
                           VertexType::InterpolationDataEvaluated;
      }

    void EvaluateCorner(VertexType &vert)
      {
        if (! (vert.mEvaluated & VertexType::ValueEvaluated))
          Evaluate(vert);
      }

    void EvaluateCenter(VertexType &vert)
      { Evaluate(vert); }

    void EvaluateElement(MeshElement_VirtualBase<VertexType> *element)
      {
        int numcorners=element->NumCorners();
        for (int i=0; i<numcorners; ++i)
          EvaluateCorner(element->Corner(i));
      }

    static ElementAction AnalyzeCubicElement(
                                MeshElement_VirtualBase<VertexType> *element,
                                ScalarType radius, VertexType *center, 
                                ScalarType epsilon, ScalarType epsilonR, 
                                const CubicErrorData &err);

    inline ElementAction AnalyzeElement(
          MeshElement_VirtualBase<VertexType> *element, 
          ScalarType radius, VertexType *center, 
          ScalarType epsilon)
      {
        return AnalyzeCubicElement(element, radius, center, 
                                   epsilon, mEpsilonR, mErr);
      }
  };


#endif