#ifndef __INTERPOLANT_2D_BASE_H
#define __INTERPOLANT_2D_BASE_H

#include "vert.h"
#include "smallmat.h"
#include "meshelt.h"
#include <assert.h>
#include "traits2d.h"

struct Extra_2D_Cubic
  {
    Matrix::SmallMatrix<ScalarType, 1, 2> mGradient;
  };

typedef VertexInfo<DomainType2d, Extra_2D_Cubic> Vertex_2D_Cubic_Base;

struct InterpStrategy
  {
    static void TriangleInterpolant(void *, 
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR);
    static void SquareInterpolant(void *, 
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR);
  };

template<class Vertex_T, class Interp>
class Interpolant2DBase 
  {
public:
    typedef Interpolant2DBase thistype;
    typedef Vertex_T VertexType;
    typedef InterpolantTraits2D<VertexType> TraitsType;
    typedef TraitsType::ContourType ContourType;
    typedef TraitsType::DomainType  DomainType;

    ScalarType mEpsilon;
    ScalarType mEpsilonR;

    virtual void EvaluateElement(MeshElement_VirtualBase<VertexType> *element) = 0;

    void InterpolateContour(std::list<ContourType> &results, 
                            MeshElementList_VirtualBase<VertexType> *elt_lst);
protected:
    ElementAction StoppingCriteriaHelper(ScalarType maxerr, 
                                ScalarType epsilon, ScalarType optimism,
                                MeshElement_VirtualBase<VertexType> *element);
    static ElementAction StoppingCriteriaHelperTriangle(ScalarType maxerr, 
                                ScalarType epsilon, ScalarType optimism,
                                MeshElement_VirtualBase<VertexType> *element);
    static ElementAction StoppingCriteriaHelperSquare(ScalarType maxerr, 
                                ScalarType epsilon, ScalarType optimism,
                                MeshElement_VirtualBase<VertexType> *element, 
                                Root2D::TriangularCubic cubics[4]);
  };

namespace Interp2D
  {
    void RenumberContours(std::list<Root2D::Contour> &contours, 
                          int mapsize, const int mappairs[][2]);

    // C0 interpolant with quadratic precision
    void NineParamater(Vertex_2D_Cubic_Base *v[3], 
                      Root2D::TriangularCubic &cubic);

    // C1 interpolant with quadratic precision
    void SibsonSplitSquare(Vertex_2D_Cubic_Base *v[4], 
                           Root2D::TriangularCubic cubics[4],
                           ScalarType epsilonR);

    // C1 interpolant with quadratic precision
    void CloughTocher(Vertex_2D_Cubic_Base *v[3], 
                      Root2D::TriangularCubic cubics[3],
                      ScalarType epsilonR);
  }

inline void InterpStrategy::TriangleInterpolant(void *, 
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR)
  {
    Interp2D::CloughTocher(v, cubics, epsilonR);
  }

inline void InterpStrategy::SquareInterpolant(void *, 
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR)
  {
    Interp2D::SibsonSplitSquare(v, cubics, epsilonR);
  }


#include "int2dbs.hpp"

#endif
