#ifndef __INTERPOLANT_2D_BASE_H
#error int2dbs.hpp should only be included from int2dbs.h
#endif

#include "merge2d.h"
#include "puzzle.h"
// begin debug code
#if ! defined(NDEBUG) && ! defined(PROFILE_HOOKS)
#define INT_2D_VIS
#include "draw.h"
#endif
// end debug code

#include "blow_profile/profiler.h"

Extern_Zone(interpolate);
Extern_Zone(contour_interpolant);
Extern_Zone(puzzle_piece);
Extern_Zone(puzzle_finish);

namespace Interp2D
  {
    void CubicHull(const Root2D::TriangularCubic &cubic,
                   ScalarType & mn, ScalarType & mx);
    void QuadraticMinAbsAvg(const Root2D::TriangularQuadraticUnit 
                             &quadratic, ScalarType & minabs,
                            ScalarType & avgabs);
  }

template<class Vertex_T, class Interp>
void Interpolant2DBase<Vertex_T, Interp>::InterpolateContour(
                                std::list<ContourType> &results, 
                                MeshElementList_VirtualBase<Vertex_T> *elt_lst)
  {
    using namespace Interp2D;
    MeshElement_VirtualBase<Vertex_T> *element;
    Puzzle puzzle(mEpsilon);

        // BEGIN DEBUG
#ifdef INT_2D_VIS
    static char filename[10];
    itoa(gDebugOutputFileCount, filename, 10);
    strcat(filename, ".eps");
    ++gDebugOutputFileCount;
    PS::Drawing drawing;
    PS::State state=drawing.GetState();
    PS::State contourstate(state);
    state.SetLineThickness(state.GetLineThickness()*.5);
    {
      PS::Path path;
      path.MoveTo(ScaleUp(0), ScaleUp(0));
      path.LineTo(ScaleUp(1), ScaleUp(0));
      path.LineTo(ScaleUp(1), ScaleUp(1));
      path.LineTo(ScaleUp(0), ScaleUp(1));
      path.Close();
      state.SetGray(0.9);
      drawing.StrokePath(path, state);
    }
    /* int elt_count=0; */
#endif
        // END DEBUG

    while ((element=elt_lst->NextElement())!=0)
      {
        assert(element->NumCorners()==3 || element->NumCorners()==4);
        VertexType *v[4];
        int i;
        int numcorners=element->NumCorners();

        EvaluateElement(element);

        for (i=0; i<numcorners; ++i)
          {
            v[i]=& element->Corner(i);
            assert(v[i]->mEvaluated == VertexType::BothEvaluated);
          }
        // Find all contours in triangle/square with vertices:
        // v[i]->mPos, v[i]->mFunctionValue, v[i]->mGradient, 

        // BEGIN DEBUG
#ifdef INT_2D_VIS
        {
          PS::Path path;
          path.MoveTo(ScaleUp(v[0]->mPos(0,0)), ScaleUp(v[0]->mPos(0,1)));
          for (i=1; i<numcorners; ++i)
            path.LineTo(ScaleUp(v[i]->mPos(0,0)), ScaleUp(v[i]->mPos(0,1)));
          path.Close();
          state.SetGray(.5);
          drawing.StrokePath(path, state);
          /*
          static char counttxt[20];
          itoa(elt_count, counttxt, 10);
          strcat(counttxt, ":");
          itoa(element->Id(), counttxt+strlen(counttxt), 10);
          DomainType2d pos=v[0]->mPos;
          for (int tmp=1; tmp<numcorners; ++tmp)
            Matrix::Add(pos, pos, v[tmp]->mPos);
          Matrix::Scale(pos, 1./numcorners);
          drawing.DrawText(counttxt, 
                      ScaleUp(pos(0,0)),
                      ScaleUp(pos(0,1)),
                      6, "Times-Roman", 
                      contourstate);
          ++elt_count;
          */
        }
#endif
        // END DEBUG

        if (numcorners==3)
          { // Clough-Tocher interpolation
            static int renum[3][6][2] = 
              { { { 0, 6 }, { 1, 0 }, { 2, 7 }, { 3, 9 }, { 4, 3 }, { 5, 4 } },
                { { 0, 7 }, { 1, 1 }, { 2, 8 }, { 3, 9 }, { 4, 4 }, { 5, 5 } },
                { { 0, 8 }, { 1, 2 }, { 2, 6 }, { 3, 9 }, { 4, 5 }, { 5, 3 } }
              };
            static int tomerge1[1] = { 7 };
            static int tomerge2[3] = { 6, 8, 9 };

            Root2D::TriangularCubic cubics[3];

            {
              Profile_Scope(interpolate);
              Interp::TriangleInterpolant(this, 
                                reinterpret_cast<Vertex_2D_Cubic_Base **>(v),
                                cubics, mEpsilonR);
            }

            typedef std::list<ContourType> ContourList;
            ContourList contours[3];

            int side;
            for (side=0; side<3; ++side)
              {
                // BEGIN DEBUG
#ifdef INT_2D_VIS
                PS::Path path;
                //ScalarType st[2];
                //
                //for (st[0]=0.0; st[0]<1.0; st[0]+=0.1)
                //  for (st[1]=0.0; st[1]<1.0-st[0]; st[1]+=0.1)
                //    {
                //      DomainType2d pos;
                //      // pos = (1-st[0]-st[1])*cubics[side].mP[0]
                //      Matrix::EqScale(pos, 1-st[0]-st[1], cubics[side].mP[0]);
                //      //pos += st[0]*cubics[side].mP[1];
                //      Matrix::AddScaled(pos, st[0], cubics[side].mP[1]);
                //      //pos += st[1]*cubics[side].mP[2];
                //      Matrix::AddScaled(pos, st[1], cubics[side].mP[2]);
                //
                //      ScalarType val = cubics[side].EvaluateAt(st);
                //      if (val<0)
                //        state.SetRGB(1,0.7,0.8);
                //      else
                //        state.SetRGB(0.7,0.8,1);
                //      path.Clear();
                //      path.Circle(ScaleUp(pos(0,0)), ScaleUp(pos(0,1)), Radius(0.01));
                //      drawing.FillPath(path, state);
                //    }
                //
                //path.Clear();
                state.SetGray(.8);
                path.MoveTo(ScaleUp(cubics[side].mP[1](0,0)), ScaleUp(cubics[side].mP[1](0,1)));
                path.LineTo(ScaleUp(cubics[side].mP[0](0,0)), ScaleUp(cubics[side].mP[0](0,1)));
                path.LineTo(ScaleUp(cubics[side].mP[2](0,0)), ScaleUp(cubics[side].mP[2](0,1)));
                drawing.StrokePath(path, state);
#endif
                // END DEBUG
                ScalarType radius=Matrix::Length(cubics[side].mP[1], 
                                                 cubics[side].mP[0]);

                Profile_Scope(contour_interpolant);
                Root2D::FindContour(cubics[side], mEpsilon/radius, mEpsilonR,
                                    contours[side]);
                RenumberContours(contours[side], 6, renum[side]);
              }

            {
              Profile_Scope(puzzle_piece);
              Merge2D::CornerNbhd corners[3];
              corners[0].mSide=9;
              corners[0].mLocation=cubics[0].mP[0];
              corners[1].mSide=4;
              corners[1].mLocation=cubics[1].mP[1];
              Merge2D::MergeContours(contours[0], contours[1], mEpsilon, 1, tomerge1, 2, corners);
              corners[1].mSide=3;
              corners[1].mLocation=cubics[0].mP[1];
              corners[2].mSide=5;
              corners[2].mLocation=cubics[2].mP[0];
              Merge2D::MergeContours(contours[0], contours[2], mEpsilon, 3, tomerge2, 3, corners);
              assert(contours[1].empty());
              assert(contours[2].empty());
              int neighbors[3];
              DomainType2d cornerpos[3];
              for (side=0; side<3; ++side)
                {
                  neighbors[side]=element->Neighbor(side);
                  cornerpos[side]=cubics[side].mP[1];
                }
              puzzle.AddPiece(results, element->Id(), 3, neighbors, contours[0], cornerpos);
            }
          }
        else
          { // Sibson split-square interpolation
            static int renum[4][6][2] = 
              { { { 0,  8 }, { 1, 0 }, { 2,  9 }, { 3, 12 }, { 4, 4 }, { 5, 5 } },
                { { 0,  9 }, { 1, 1 }, { 2, 10 }, { 3, 12 }, { 4, 5 }, { 5, 6 } },
                { { 0, 10 }, { 1, 2 }, { 2, 11 }, { 3, 12 }, { 4, 6 }, { 5, 7 } },
                { { 0, 11 }, { 1, 3 }, { 2,  8 }, { 3, 12 }, { 4, 7 }, { 5, 4 } }
              };
            static int tomerge1[1] = { 9 };
            static int tomerge2[1] = { 10 };
            static int tomerge3[3] = { 8, 11, 12 };

            Root2D::TriangularCubic cubics[4];

            {
              Profile_Scope(interpolate);
              Interp::SquareInterpolant(this, 
                                reinterpret_cast<Vertex_2D_Cubic_Base **>(v),
                                cubics, mEpsilonR);
            }

            typedef std::list<ContourType> ContourList;
            ContourList contours[4];

            int side;
            for (side=0; side<4; ++side)
              {
                // BEGIN DEBUG
#ifdef INT_2D_VIS
                PS::Path path;
                //ScalarType st[2];
                //
                //for (st[0]=0.0; st[0]<1.0; st[0]+=0.1)
                //  for (st[1]=0.0; st[1]<1.0-st[0]; st[1]+=0.1)
                //    {
                //      DomainType2d pos;
                //      // pos = (1-st[0]-st[1])*cubics[side].mP[0]
                //      Matrix::EqScale(pos, 1-st[0]-st[1], cubics[side].mP[0]);
                //      //pos += st[0]*cubics[side].mP[1];
                //      Matrix::AddScaled(pos, st[0], cubics[side].mP[1]);
                //      //pos += st[1]*cubics[side].mP[2];
                //      Matrix::AddScaled(pos, st[1], cubics[side].mP[2]);
                //
                //      ScalarType val = cubics[side].EvaluateAt(st);
                //      if (val<0)
                //        state.SetRGB(1,0.7,0.8);
                //      else
                //        state.SetRGB(0.7,0.8,1);
                //      path.Clear();
                //      path.Circle(ScaleUp(pos(0,0)), ScaleUp(pos(0,1)), Radius(0.01));
                //      drawing.FillPath(path, state);
                //    }
                //
                //path.Clear();
                state.SetGray(.8);
                path.MoveTo(ScaleUp(cubics[side].mP[1](0,0)), ScaleUp(cubics[side].mP[1](0,1)));
                path.LineTo(ScaleUp(cubics[side].mP[0](0,0)), ScaleUp(cubics[side].mP[0](0,1)));
                path.LineTo(ScaleUp(cubics[side].mP[2](0,0)), ScaleUp(cubics[side].mP[2](0,1)));
                drawing.StrokePath(path, state);
#endif
                // END DEBUG
                ScalarType radius=Matrix::Length(cubics[side].mP[1], 
                                                 cubics[side].mP[0]);

                Profile_Scope(contour_interpolant);
                Root2D::FindContour(cubics[side], mEpsilon/radius, mEpsilonR,
                                    contours[side]);

                RenumberContours(contours[side], 6, renum[side]);
              }

            {
              Profile_Scope(puzzle_piece);
              Merge2D::CornerNbhd corners[3];
              corners[0].mSide=12;
              corners[0].mLocation=cubics[0].mP[0];
              corners[1].mSide=5;
              corners[1].mLocation=cubics[1].mP[1];
              Merge2D::MergeContours(contours[0], contours[1], mEpsilon, 1, tomerge1, 2, corners);
              corners[1].mSide=6;
              corners[1].mLocation=cubics[2].mP[1];
              Merge2D::MergeContours(contours[0], contours[2], mEpsilon, 1, tomerge2, 2, corners);
              corners[1].mSide=7;
              corners[1].mLocation=cubics[3].mP[1];
              corners[2].mSide=4;
              corners[2].mLocation=cubics[0].mP[1];
              Merge2D::MergeContours(contours[0], contours[3], mEpsilon, 3, tomerge3, 3, corners);
              assert(contours[1].empty());
              assert(contours[2].empty());
              assert(contours[3].empty());
              int neighbors[4];
              DomainType2d cornerpos[4];
              for (side=0; side<4; ++side)
                {
                  neighbors[side]=element->Neighbor(side);
                  cornerpos[side]=cubics[side].mP[1];
                }
              puzzle.AddPiece(results, element->Id(), 4, neighbors, contours[0], cornerpos);
            }
          }
      }
    {
      Profile_Scope(puzzle_finish);
      puzzle.FinishPuzzle(results);
    }
        // BEGIN DEBUG
#ifdef INT_2D_VIS
    DrawSplineContour(results, drawing, contourstate);
    // draw ideal
    //state.SetRGB(0,1,0);
    //  {
    //    PS::Path path;
    //    path.Circle(ScaleUp(.25), ScaleUp(.25), Radius(.2));
    //    drawing.StrokePath(path, state);
    //  }
    WriteEPSFile(drawing, filename, "InterpolateContour", "Shape");
#endif
        // END DEBUG
  }

template<class Vertex_T, class Interp>
ElementAction Interpolant2DBase<Vertex_T, Interp>
::StoppingCriteriaHelper(ScalarType maxerr, ScalarType epsilon, 
                         ScalarType optimism, /* 0..1 */
                         MeshElement_VirtualBase<VertexType> *element)
  {
    using namespace Interp2D;
    assert(element->NumCorners()==3 || element->NumCorners()==4);
    VertexType *v[4];
    int i;
    int numcorners=element->NumCorners();

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
      }
    // Find all contours in triangle/square with vertices:
    // v[i]->mPos, v[i]->mFunctionValue, v[i]->mGradient, 

    ScalarType minabsgradient, avgabsgradient;

    if (numcorners==3)
      { // could use either 9 parameter (quadratic-precision) interpolant
        // or Clough-Tocher; much easier to figure out gradients
        // for 9 parameter interpolant...
        Root2D::TriangularCubic cubic;

        NineParamater(reinterpret_cast<Vertex_2D_Cubic_Base **>(v), cubic);
        
        ScalarType mn, mx;
        CubicHull(cubic, mn, mx);

        if ((mn-maxerr)>0 || (mx+maxerr)<0)
          return DiscardElement;

        Root2D::TriangularQuadraticUnit gradients[2];
        cubic.PartialDerivativeS(gradients[0]);
        cubic.PartialDerivativeT(gradients[1]);

        minabsgradient=0;
        avgabsgradient=0;
        for (i=0; i<2; ++i)
          {
            ScalarType absgradmin;
            ScalarType absgradavg;
            QuadraticMinAbsAvg(gradients[i], absgradmin, absgradavg);
            minabsgradient += absgradmin*absgradmin;
            avgabsgradient += absgradavg*absgradavg;
          }
        minabsgradient=sqrt(minabsgradient);
        avgabsgradient=sqrt(avgabsgradient);
      }
    else
      { // numcorners==4
        Root2D::TriangularCubic cubics[4];

        SibsonSplitSquare(reinterpret_cast<Vertex_2D_Cubic_Base **>(v),
                          cubics, mEpsilonR);

        ScalarType mn, mx;
        CubicHull(cubics[0], mn, mx);
        int j;
        for (j=1; j<4; ++j)
          {
            ScalarType mn_temp, mx_temp;
            CubicHull(cubics[j], mn_temp, mx_temp);
            if (mn_temp<mn)
              mn=mn_temp;
            if (mx<mx_temp)
              mx=mx_temp;
          }

        if ((mn-maxerr)>0 || (mx+maxerr)<0)
          return DiscardElement;

        avgabsgradient=0;
        for (j=0; j<4; ++j)
          {
            Root2D::TriangularQuadraticUnit gradients[2];
            cubics[j].PartialDerivativeS(gradients[0]);
            cubics[j].PartialDerivativeT(gradients[1]);

            ScalarType curminabsgrad=0, curavgabsgrad=0;
            for (i=0; i<2; ++i)
              {
                ScalarType absgradmin;
                ScalarType absgradavg;
                QuadraticMinAbsAvg(gradients[i], absgradmin, absgradavg);
                curminabsgrad += absgradmin*absgradmin;
                curavgabsgrad += absgradavg*absgradavg;
              }

            avgabsgradient+=sqrt(curavgabsgrad);

            if (j==0 || curminabsgrad<minabsgradient)
              minabsgradient = curminabsgrad;
          }
        minabsgradient=sqrt(minabsgradient);
        avgabsgradient*=0.25;
      }

    ScalarType absgrad=minabsgradient+(avgabsgradient-minabsgradient)*optimism;
    return (epsilon*absgrad > maxerr) ? ContourElement : RefineElement;
  }

template<class Vertex_T, class Interp>
ElementAction Interpolant2DBase<Vertex_T, Interp>
::StoppingCriteriaHelperTriangle(ScalarType maxerr, ScalarType epsilon, 
                         ScalarType optimism, /* 0..1 */
                         MeshElement_VirtualBase<VertexType> *element)
  {
    using namespace Interp2D;
    assert(element->NumCorners()==3);
    VertexType *v[3];
    int i;
    const int numcorners=3;

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
      }
    // Find all contours in triangle/square with vertices:
    // v[i]->mPos, v[i]->mFunctionValue, v[i]->mGradient, 

    ScalarType minabsgradient, avgabsgradient;

    // could use either 9 parameter (quadratic-precision) interpolant
    // or Clough-Tocher; much easier to figure out gradients
    // for 9 parameter interpolant...
    Root2D::TriangularCubic cubic;

    NineParamater(reinterpret_cast<Vertex_2D_Cubic_Base **>(v), cubic);
    
    ScalarType mn, mx;
    CubicHull(cubic, mn, mx);

    if ((mn-maxerr)>0 || (mx+maxerr)<0)
      return DiscardElement;

    Root2D::TriangularQuadraticUnit gradients[2];
    cubic.PartialDerivativeS(gradients[0]);
    cubic.PartialDerivativeT(gradients[1]);

    minabsgradient=0;
    avgabsgradient=0;
    for (i=0; i<2; ++i)
      {
        ScalarType absgradmin;
        ScalarType absgradavg;
        QuadraticMinAbsAvg(gradients[i], absgradmin, absgradavg);
        minabsgradient += absgradmin*absgradmin;
        avgabsgradient += absgradavg*absgradavg;
      }
    minabsgradient=sqrt(minabsgradient);
    avgabsgradient=sqrt(avgabsgradient);
    ScalarType absgrad=minabsgradient+(avgabsgradient-minabsgradient)*optimism;
    return (epsilon*absgrad > maxerr) ? ContourElement : RefineElement;
  }

template<class Vertex_T, class Interp>
ElementAction Interpolant2DBase<Vertex_T, Interp>
::StoppingCriteriaHelperSquare(ScalarType maxerr, ScalarType epsilon, 
                         ScalarType optimism, /* 0..1 */
                         MeshElement_VirtualBase<VertexType> *element,
                         Root2D::TriangularCubic cubics[4])
  {
    using namespace Interp2D;
    assert(element->NumCorners()==4);
    VertexType *v[4];
    int i;
    const int numcorners=4;

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
      }
    // Find all contours in triangle/square with vertices:
    // v[i]->mPos, v[i]->mFunctionValue, v[i]->mGradient, 

    ScalarType minabsgradient, avgabsgradient;
    ScalarType mn, mx;
    CubicHull(cubics[0], mn, mx);
    int j;
    for (j=1; j<4; ++j)
      {
        ScalarType mn_temp, mx_temp;
        CubicHull(cubics[j], mn_temp, mx_temp);
        if (mn_temp<mn)
          mn=mn_temp;
        if (mx<mx_temp)
          mx=mx_temp;
      }

    if ((mn-maxerr)>0 || (mx+maxerr)<0)
      return DiscardElement;

    avgabsgradient=0;
    for (j=0; j<4; ++j)
      {
        Root2D::TriangularQuadraticUnit gradients[2];
        cubics[j].PartialDerivativeS(gradients[0]);
        cubics[j].PartialDerivativeT(gradients[1]);

        ScalarType curminabsgrad=0, curavgabsgrad=0;
        for (i=0; i<2; ++i)
          {
            ScalarType absgradmin;
            ScalarType absgradavg;
            QuadraticMinAbsAvg(gradients[i], absgradmin, absgradavg);
            curminabsgrad += absgradmin*absgradmin;
            curavgabsgrad += absgradavg*absgradavg;
          }

        avgabsgradient+=sqrt(curavgabsgrad);

        if (j==0 || curminabsgrad<minabsgradient)
          minabsgradient = curminabsgrad;
      }
    minabsgradient=sqrt(minabsgradient);
    avgabsgradient*=0.25;

    ScalarType absgrad=minabsgradient+(avgabsgradient-minabsgradient)*optimism;
    return (epsilon*absgrad > maxerr) ? ContourElement : RefineElement;
  }
