#ifndef __INTERPOLANT_LINEAR_2D_H
#define __INTERPOLANT_LINEAR_2D_H

#include "template.h"
#include "vert.h"
#include "traits2d.h"
#include "func2d.h"
#include "meshelt.h"

struct Extra_2D_Linear
  {
  };

class InterpolantLinear2D
  {
public:
    typedef VertexInfo<DomainType2d, Extra_2D_Linear> VertexType;
    typedef InterpolantTraits2D<VertexType> TraitsType;
    typedef TraitsType::ContourType ContourType;

    typedef FunctionLinear2D FunctionType;
    FunctionType *mFunction;

    ScalarType mConservativeError;
    ScalarType mThirdOrderError;
    ScalarType mEpsilon;
    ScalarType mEpsilonR;

    void Evaluate(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::ValueEvaluated));
        vert.mFunctionValue=mFunction->Value(vert.mPos(0,0), vert.mPos(0,1));
        vert.mEvaluated |= VertexType::ValueEvaluated |
                           VertexType::InterpolationDataEvaluated;
      }

    void EvaluateCorner(VertexType &vert)
      {
        if (! (vert.mEvaluated & VertexType::ValueEvaluated))
          Evaluate(vert);
      }

    void EvaluateCenter(VertexType &vert)
      {
        Evaluate(vert);
      }

    ElementAction AnalyzeElement(MeshElement_VirtualBase<VertexType> *element, 
                                 ScalarType radius, VertexType *center, 
                                 ScalarType epsilon);

    void InterpolateContour(std::list<ContourType> &results, 
                            MeshElementList_VirtualBase<VertexType> *elt_lst);
  };

#endif
