/* int2dmd.h */
#ifndef __INTERPOLANT_2D_MAXD_H
#define __INTERPOLANT_2D_MAXD_H

#include "int2dbs.h"
#include "func2dmd.h"

class Interpolant2DMaxD 
: public Interpolant2DBase<VertexInfoMaxD<DomainType2d, Extra_2D_Cubic>,
                           InterpStrategy>
  {
public:
    ScalarType mOptimism;
    ScalarType mErrorBound;
    Function2DMaxD *mFunction;

    void EvaluateValue(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::ValueEvaluated));
        ScalarType x=vert.mPos(0,0);
        ScalarType y=vert.mPos(0,1);
        vert.mFunctionValue=mFunction->Value(x,y);
        vert.mEvaluated |= VertexType::ValueEvaluated;
      }
    void EvaluateInterpolationData(VertexType &vert)
      {
        assert(! (vert.mEvaluated & VertexType::InterpolationDataEvaluated));
        mFunction->Grad(vert.mPos(0,0), vert.mPos(0,1),
                        vert.mGradient(0,0), vert.mGradient(0,1)); 
        vert.mEvaluated |= VertexType::InterpolationDataEvaluated;
      }

    void EvaluateElement(MeshElement_VirtualBase<VertexType> *element)
      {
        VertexType *v[4];
        int numcorners=element->NumCorners();
        for (int i=0; i<numcorners; ++i)
          {
            v[i]=& element->Corner(i);
            if (v[i]->mEvaluated != VertexType::BothEvaluated)
              {
                if (!(v[i]->mEvaluated & VertexType::ValueEvaluated))
                  EvaluateValue(*v[i]);
                EvaluateInterpolationData(*v[i]);
              }
          }
      }

    ScalarType SafeRadius(VertexType &vert, ScalarType maxRadius)
      {
        assert(maxRadius>0);
        if (! (vert.mEvaluated & VertexType::ValueEvaluated))
          EvaluateValue(vert);

        ScalarType x=vert.mPos(0,0);
        ScalarType y=vert.mPos(0,1);
        
        // determine max derivative
        ScalarType maxderivative=mFunction->MaxDeriv(x,y,maxRadius);
        return SafeRadiusHelper(vert.mFunctionValue, maxderivative, maxRadius);
      }
    bool StoppingCriteria(MeshElement_VirtualBase<VertexType> *element, 
                          ScalarType radius, ScalarType epsilon)
      {
        for (int i=0; i<element->NumCorners(); ++i)
          {
            VertexType *v=& element->Corner(i);
            assert(v->mEvaluated & VertexType::ValueEvaluated);
            if (! (v->mEvaluated & VertexType::InterpolationDataEvaluated))
              EvaluateInterpolationData(*v);
          }


        ScalarType maxerr;
        if (element->NumCorners()==3)
          maxerr=TriangleMaxError(radius, mErrorBound);
        else
          maxerr=SquareMaxError(radius, mErrorBound);
        return StoppingCriteriaHelper(maxerr, epsilon, mOptimism, element)
                != RefineElement;
      }

protected:
    ScalarType SafeRadiusHelper(ScalarType value, ScalarType maxderiv, ScalarType maxRadius)
      {
        ScalarType maxsafe = fabs(value/maxderiv);
        return Matrix::Min(maxsafe,maxRadius);
      }

    ScalarType SquareMaxError(ScalarType radius, ScalarType maxf3)
      {
        // compute radius^3
        ScalarType radius3 = radius * radius * radius; 
        return radius3 * maxf3 * .04554899040;
      }

    ScalarType TriangleMaxError(ScalarType radius, ScalarType maxf3)
      {
        // compute radius^3
        ScalarType radius3 = radius * radius * radius; 
        return radius3 * maxf3 * .03183055317;
      }
  };

#endif