#include "merge2d.h"
#include "fixme.h"
#include <hash_map>

namespace Merge2D
  {
    class PrivateImpl
      {
        typedef std::vector<DomainType2d> ControlVect;
        typedef ControlVect::iterator     ControlIter;
        typedef ControlVect::reverse_iterator ControlRIter;

        typedef Root2D::Contour  ContourType;
        typedef std::list<ContourType> ContourList;
        typedef ContourList::iterator  ContourIter;

    public:
        struct End
          {
            ScalarType  mPseudoAngle;
            int         mDestSrc;
            ContourIter mIter;
            int         mFirstLast;
            bool        mShort;
          };
    private:

        typedef std::list<End>    EndList;
        typedef EndList::iterator EndIter;

        struct Star
          {
            int mSide;
            DomainType2d mBounds[2];
            EndList mEnds;
          };

        typedef std::list<Star>    StarList;
        typedef StarList::iterator StarIter;

        //typedef std::list<StarIter>     StarIterList;
        //typedef StarIterList::iterator  StarIterIter;

        struct EndId
          {
            ContourType *mContour;
            int mFirstLast;
          };

        struct EndIdEqual : public std::binary_function<EndId, EndId, bool>
          {
            bool operator()(const EndId &l, const EndId &r) const
              {
                return l.mContour==r.mContour && l.mFirstLast==r.mFirstLast;
              }
          };

        struct EndHashFunc : public std::unary_function<EndId, size_t>
          {
            size_t operator()(const EndId& e) const
              {
                return reinterpret_cast<size_t>(e.mContour) + 
                       static_cast<size_t>(e.mFirstLast);
              }
          };

        struct StarId
          {
            StarIter mStar;
            EndIter mStrand;
          };

        typedef std::hash_map<EndId, StarId, EndHashFunc, EndIdEqual> EndMap;
        typedef EndMap::value_type EndMapPair;
        typedef EndMap::iterator EndMapIter;

        StarList mStars;
        //StarIterList mWhichStar;
        EndMap mEndMap;

        static DomainType2d GetEndPos(const End & end)
          {
            if (end.mFirstLast)
              return (*end.mIter).mControl.back();
            else
              return (*end.mIter).mControl.front();
          }

        void AddToStar(StarIter s, int srcdest, ContourIter src_i, int firstlast, ScalarType epsilon);

        void CreateStars(ContourList &contours, int srcdest, ScalarType epsilon, 
                         int numtomerge, int *tomerge);

        bool CheckCorners(int numcorners, CornerNbhd *corners, ScalarType epsilon);

        void ConnectStarValenceTwo(ContourList &dest, ContourList &source);

        void ConnectStars(ContourList &contours);

#ifndef NDEBUG
        void CheckEndMap();
#endif

        static void AttachStrands(ContourType *dest, ContourType *src,
                                  int dst_end, int src_end, 
                                  bool dst_short, bool src_short);
        static bool FindDistantPoint(const std::vector<DomainType2d> &toscan, 
                              int firstlast, ScalarType epsilon, 
                              DomainType2d &endpt, DomainType2d &nextpt);
        static bool CloseToBox(DomainType2d box[2], DomainType2d pt, 
                        ScalarType dist);

    public:
        static void DoMerging(ContourList &dest, ScalarType epsilon, 
          int numtomerge, int *tomerge, int numcorners, CornerNbhd *corners);

        static void DoConnecting(ContourList &dest, ContourList &source, 
                       ScalarType epsilon, int numtomerge, int *tomerge);
      };

    
    static inline bool operator<(const PrivateImpl::End &l, 
                                 const PrivateImpl::End &r)
      {
        return l.mPseudoAngle < r.mPseudoAngle;
      }

#ifndef NDEBUG
    void PrivateImpl::CheckEndMap()
      {
        for (EndMapIter i=mEndMap.begin(); i!=mEndMap.end(); ++i)
          {
            *(*i).second.mStar;
            *(*i).second.mStrand;
          }
      }
#endif
    
    void PrivateImpl ::
    AttachStrands(ContourType *dest, ContourType *src, 
                  int dst_end, int src_end, 
                  bool dst_short, bool src_short)
      {
        assert(dest->mEndingSides[dst_end]==src->mEndingSides[src_end]);
        dest->mEndingSides[dst_end]=src->mEndingSides[1-src_end];
        if (dst_short)
          {
            DomainType2d dst_pt;
            if (dst_end)
              dst_pt=dest->mControl.front();
            else
              dst_pt=dest->mControl.back();
            if (dst_end==src_end)
              std::reverse(src->mControl.begin(), src->mControl.end());
            dest->mControl.swap(src->mControl);
            if (dst_end)
              dest->mControl.front()=dst_pt;
            else
              dest->mControl.back()=dst_pt;
            src->mControl.clear();
            return;
          }
        if (src_short)
          {
            DomainType2d src_pt;
            if (src_end)
              src_pt=src->mControl.front();
            else
              src_pt=src->mControl.back();
            if (dst_end)
              dest->mControl.back()=src_pt;
            else
              dest->mControl.front()=src_pt;
            return;
          }
#pragma fixme("do fixup for unknown-at-edge case")
#pragma fixme("adjust endpoints to be coincident")
        if (dst_end==1)
          if (src_end==0)
            {
              dest->mControl.reserve(dest->mControl.size()+src->mControl.size());
              ControlIter src_i=src->mControl.begin();
              for (++src_i; src_i!=src->mControl.end(); ++src_i)
                dest->mControl.push_back(*src_i);
              src->mControl.clear();
            }
          else // (src_end==1)
            {
              dest->mControl.reserve(dest->mControl.size()+src->mControl.size());
              ControlRIter src_i=src->mControl.rbegin();
              for (++src_i; src_i!=src->mControl.rend(); ++src_i)
                dest->mControl.push_back(*src_i);
              src->mControl.clear();
            }
        else // (dst_end==0)
          {
            if (src_end==0)
              std::reverse(src->mControl.begin(), src->mControl.end());
            src->mControl.reserve(dest->mControl.size()+src->mControl.size());
            ControlIter dst_i=dest->mControl.begin();
            for (++dst_i; dst_i!=dest->mControl.end(); ++dst_i)
              src->mControl.push_back(*dst_i);
            dest->mControl.swap(src->mControl);
            src->mControl.clear();
          }
      }

    ScalarType PseudoAngle(ScalarType dx, ScalarType dy)
      {
        if (dx>dy)
          {
            if (dx>-dy)
              return 1+(dy/dx);
            else if (dx<-dy)
              return 7-dx/dy;
            else
              return 0;
          }
        else if (dx<dy)
          {
            if (dx>-dy)
              return 3-(dx/dy);
            else if (dx<-dy)
              return 5+dy/dx;
            else
              return 4;
          }
        else if (dx>0)
          return 2;
        return 6;
      }

    
    bool PrivateImpl ::
    FindDistantPoint(const std::vector<DomainType2d> &toscan, 
                          int firstlast, ScalarType epsilon, 
                          DomainType2d &endpt, DomainType2d &nextpt)
      {
        // (record first point > epsilon away from start if so)
        typedef std::vector<DomainType2d> ControlVect;
        typedef ControlVect::const_iterator ControlConstIter;
        if (firstlast)
          { // last, go backwards through strand
            ControlConstIter next_i=toscan.end();
            assert(next_i!=toscan.begin());
            --next_i;
            endpt=*next_i;
            assert(next_i!=toscan.begin());
            --next_i;
            while (1)
              {
                nextpt= *next_i;
                if (Matrix::Length(endpt, nextpt)>=epsilon)
                  return true;
                if (next_i==toscan.begin())
                  return false;
                --next_i;
              }
          }
        else
          { // first, go forwards through strand
            ControlConstIter next_i=toscan.begin();
            assert(next_i!=toscan.end());
            endpt=*next_i;
            ++next_i;
            assert(next_i!=toscan.end());
            while (1)
              {
                nextpt= *next_i;
                if (Matrix::Length(endpt, nextpt)>=epsilon)
                  return true;
                ++next_i;
                if (next_i==toscan.end())
                  return false;
              }
          }
      }

    
    bool PrivateImpl ::
    CloseToBox(DomainType2d box[2], DomainType2d pt, ScalarType dist)
      {
        DomainType2d delta;
        if (pt(0,0)<box[0](0,0))
          delta(0,0)=box[0](0,0)-pt(0,0);
        else if (pt(0,0)>box[1](0,0))
          delta(0,0)=pt(0,0)-box[1](0,0);
        else
          delta(0,0)=0;
        if (pt(0,1)<box[0](0,1))
          delta(0,1)=box[0](0,1)-pt(0,1);
        else if (pt(0,1)>box[1](0,1))
          delta(0,1)=pt(0,1)-box[1](0,1);
        else
          delta(0,1)=0;
        // is pt within dist of the box?
        if (delta(0,0)+delta(0,1) > 2*dist ||
            Matrix::TwoNorm(delta) > dist)
          return false; // no
        return true;
      }

    void EnlargeBox(DomainType2d box[2], DomainType2d pt)
      {
        if (pt(0,0) < box[0](0,0))
          box[0](0,0) = pt(0,0);
        else if (pt(0,0)>box[1](0,0))
          box[1](0,0) = pt(0,0);
        if (pt(0,1)<box[0](0,1))
          box[0](0,1) = pt(0,1);
        else if (pt(0,1)>box[1](0,1))
          box[1](0,1) = pt(0,1);
      }

    
    void PrivateImpl ::
    AddToStar(StarIter s, int srcdest, ContourIter src_i, int firstlast, ScalarType epsilon)
      {
        // create new end in star
        End end;

        // find a point of *src_i > epsilon from endpt
        DomainType2d endpt, nextpt;
        end.mShort =
          ! FindDistantPoint((*src_i).mControl, firstlast,
                             epsilon/* *(1./16) */, endpt, nextpt);

        end.mPseudoAngle=PseudoAngle(nextpt(0,0)-endpt(0,0),
                                     nextpt(0,1)-endpt(0,1));
        end.mDestSrc=srcdest;
        end.mIter=src_i;
        end.mFirstLast=firstlast;

        EndIter ei=(*s).mEnds.begin();
        while (ei!=(*s).mEnds.end() && (*ei)<end)
          ++ei;
        ei=(*s).mEnds.insert(ei, end);
        //(*s).mEnds.push_back();
        //End &end=(*s).mEnds.back();

        EndId endid;
        endid.mContour = &(*end.mIter);
        endid.mFirstLast = firstlast;

        StarId starid;
        starid.mStar = s;
        starid.mStrand = ei;

        mEndMap.insert(EndMapPair(endid, starid));
#ifndef NDEBUG
        CheckEndMap();
#endif
      }

    
    void PrivateImpl ::
    CreateStars(ContourList &contours, int srcdest, ScalarType epsilon, 
                         int numtomerge, int *tomerge)
      {
        for (ContourIter src_i=contours.begin(); 
              src_i!=contours.end(); ++src_i)
          {
            for (int firstlast=0; firstlast<2; ++firstlast)
              {
                // look up src_side in tomerge[]
                bool found=false;
                int src_side=(*src_i).mEndingSides[firstlast];
                for (int i=0; i<numtomerge; ++i)
                  if (src_side==tomerge[i])
                    {
                      found=true;
                      break;
                    }
                if (! found)
                  {
                    continue; // ignore this src_side
                  }
                // have a src_side that we are supposed to try to merge
                // look for an existing star
                DomainType2d endpt;
                if (firstlast)
                  endpt=(*src_i).mControl.back();
                else
                  endpt=(*src_i).mControl.front();
                found=false;
                for (StarIter s=mStars.begin(); s!=mStars.end(); ++s)
                  if ((*s).mSide==src_side)
                    {
                      // compute the distance between endpt and (*s)'s 
                      // bounding box
                      // is endpt within epsilon of the star *s?
                      if (! CloseToBox((*s).mBounds, endpt, epsilon))
                        continue; // no, check next star
                      // yes! add this end to the star
                      found=true;

                      // enlarge bounding box
                      EnlargeBox((*s).mBounds, endpt);

                      AddToStar(s, srcdest, src_i, firstlast, epsilon);
                      break;
                    }
                if (! found)
                  { // did not find a star to add it to, create a new one!
                    mStars.push_back();
                    mStars.back().mSide=src_side;
                    mStars.back().mBounds[0]=endpt;
                    mStars.back().mBounds[1]=endpt;

                    AddToStar(--mStars.end(), srcdest, src_i, 
                              firstlast, epsilon);
                  }
              }
          }
      }

    void PrivateImpl ::
    ConnectStarValenceTwo(ContourList &dest, ContourList &source)
      {
#ifndef NDEBUG
        {
          for (StarIter s=mStars.begin(); s!=mStars.end(); ++s)
            {
              assert((*s).mEnds.size()==2);
              assert((*s).mEnds.front().mDestSrc!=(*s).mEnds.back().mDestSrc);
            }
        }
#endif
        while (! mStars.empty())
          {
            StarIter star[2];

            star[0]=mStars.begin();
            assert((*star[0]).mEnds.size()==2);
            assert((*star[0]).mEnds.front().mDestSrc!=
                   (*star[0]).mEnds.back().mDestSrc);

            EndIter src_end[2];
            EndIter dst_end[2];
            if ((*star[0]).mEnds.front().mDestSrc==1)
              {
                src_end[0]=(*star[0]).mEnds.begin();
                dst_end[0]=src_end[0];
                ++dst_end[0];
              }
            else
              {
                dst_end[0]=(*star[0]).mEnds.begin();
                src_end[0]=dst_end[0];
                ++src_end[0];
              }

            EndId endid;
            endid.mContour=& (*(*src_end[0]).mIter);
            endid.mFirstLast=1-(*src_end[0]).mFirstLast;
            EndMapIter iter=mEndMap.find(endid);
            if (iter==mEndMap.end())
              { // connect only one star
                AttachStrands(&(*(*dst_end[0]).mIter),
                              &(*(*src_end[0]).mIter),
                              (*dst_end[0]).mFirstLast, 
                              (*src_end[0]).mFirstLast,
                              (*dst_end[0]).mShort, 
                              (*src_end[0]).mShort);
                //endid.mFirstLast=(*src_end[0]).mFirstLast;
                //mEndMap.erase(endid);
                //endid.mContour=& (*(*dst_end[0]).mIter);
                //endid.mFirstLast=(*dst_end[0]).mFirstLast;
                //mEndMap.erase(endid);
                source.erase((*src_end[0]).mIter);
                mStars.erase(star[0]);
              }
            else
              { // connect both ends of src_end[0]
                star[1]=(*iter).second.mStar;
                assert((*star[1]).mEnds.size()==2);
                assert((*star[1]).mEnds.front().mDestSrc!=
                       (*star[1]).mEnds.back().mDestSrc);
                src_end[1]=(*iter).second.mStrand;
                assert((*src_end[0]).mFirstLast != (*src_end[1]).mFirstLast);
                dst_end[1]=(*star[1]).mEnds.begin();
                if (dst_end[1]==src_end[1])
                  ++dst_end[1];
                if ((*dst_end[0]).mIter == (*dst_end[1]).mIter)
                  { // form a loop!
                    AttachStrands(&(*(*dst_end[0]).mIter),
                                  &(*(*src_end[0]).mIter),
                                  (*dst_end[0]).mFirstLast,
                                  (*src_end[0]).mFirstLast,
                                  (*dst_end[0]).mShort, 
                                  (*src_end[0]).mShort);
                    (*(*dst_end[0]).mIter).mEndingSides[0] = -1;
                    (*(*dst_end[0]).mIter).mEndingSides[1] = -1;
                  }
                else
                  { // connect three strands
                    AttachStrands(&(*(*dst_end[0]).mIter),
                                  &(*(*src_end[0]).mIter),
                                  (*dst_end[0]).mFirstLast,
                                  (*src_end[0]).mFirstLast,
                                  (*dst_end[0]).mShort, 
                                  (*src_end[0]).mShort);
                    AttachStrands(&(*(*dst_end[0]).mIter),
                                  &(*(*dst_end[1]).mIter),
                                  (*dst_end[0]).mFirstLast,
                                  (*dst_end[1]).mFirstLast,
                                  (*dst_end[0]).mShort &&
                                  (*src_end[0]).mShort,
                                  (*dst_end[1]).mShort);
                    // erase one element of dest
                    dest.erase((*dst_end[1]).mIter);
                  }
                // erase one element of source
                source.erase((*src_end[0]).mIter);
                // erase two stars
                mStars.erase(star[0]);
                mStars.erase(star[1]);
                // should also erase 4 entries from mEndMap
              }
          }
      }

    
    void PrivateImpl ::
    ConnectStars(ContourList &contours)
      {
        for (StarIter s=mStars.begin(); s!=mStars.end(); ++s)
          {
#ifndef NDEBUG
            CheckEndMap();
#endif
            while ((*s).mEnds.size()>1)
              {
                // decide how to pair up ends to maximize corner angle---
                EndIter cur[2];
                EndIter best[2];
                best[0]=cur[0]=cur[1]=(*s).mEnds.begin();
                ++cur[1];
                best[1]=cur[1];
                // ideal is an angle difference close to 4 (equivalent to pi)
                bool best_shortdiff=((*cur[0]).mShort != (*cur[1]).mShort);
                ScalarType best_v;
                if (best_shortdiff)
                  best_v=Matrix::Length(GetEndPos(*cur[0]), GetEndPos(*cur[1]));
                else
                  best_v=Matrix::Abs(4-(*cur[1]).mPseudoAngle
                                      +(*cur[0]).mPseudoAngle);
                ScalarType cur_v;
                bool cur_shortdiff;
                while (1)
                  {
                    cur[0]=cur[1];
                    ++cur[1];
                    if (cur[1]==(*s).mEnds.end())
                      break;
                    cur_shortdiff=((*cur[0]).mShort != (*cur[1]).mShort);
                    if (cur_shortdiff)
                      {
                        cur_v=Matrix::Length(GetEndPos(*cur[0]), GetEndPos(*cur[1]));
                        if (! best_shortdiff || cur_v<best_v)
                          {
                            best_shortdiff=true;
                            best_v=cur_v;
                            best[0]=cur[0];
                            best[1]=cur[1];
                          }
                      }
                    else if (! best_shortdiff)
                      {
                        cur_v=Matrix::Abs(4-(*cur[1]).mPseudoAngle
                                          +(*cur[0]).mPseudoAngle);
                        if (cur_v<best_v)
                          {
                            best_v=cur_v;
                            best[0]=cur[0];
                            best[1]=cur[1];
                          }
                      }
                  }
                cur[1]=(*s).mEnds.begin();
                cur_shortdiff=((*cur[0]).mShort != (*cur[1]).mShort);

                if (cur_shortdiff)
                  {
                    cur_v=Matrix::Length(GetEndPos(*cur[0]), GetEndPos(*cur[1]));
                    if (! best_shortdiff || cur_v<best_v)
                      {
                        best[0]=cur[0];
                        best[1]=cur[1];
                      }
                  }
                else if (! best_shortdiff)
                  {
                    cur_v=Matrix::Abs(4+(*cur[1]).mPseudoAngle
                                      -(*cur[0]).mPseudoAngle);
                    if (cur_v<best_v)
                      {
                        best[0]=cur[0];
                        best[1]=cur[1];
                      }
                  }

                // want to connect best[0] to best[1]
                EndId endid;
                if ((*best[0]).mIter==(*best[1]).mIter)
                  { // form a loop
                    assert((*best[0]).mFirstLast != (*best[1]).mFirstLast);
                    (*(*best[0]).mIter).mEndingSides[0] = -1;
                    (*(*best[0]).mIter).mEndingSides[1] = -1;
                    // update mEndMap
                    endid.mContour = & (*(*best[0]).mIter);
                    endid.mFirstLast = 0;
                    mEndMap.erase(endid);
                    endid.mFirstLast = 1;
                    mEndMap.erase(endid);
                  }
                else
                  { // connect two strands together
                    AttachStrands(&(*(*best[0]).mIter),
                                  &(*(*best[1]).mIter),
                                  (*best[0]).mFirstLast,
                                  (*best[1]).mFirstLast,
                                  (*best[0]).mShort, 
                                  (*best[1]).mShort);
                    // clean up other ends and update mEndMap
                    bool still_short = (*best[0]).mShort && 
                                       (*best[1]).mShort;
                    endid.mContour = & (*(*best[0]).mIter);
                    endid.mFirstLast = 1-(*best[0]).mFirstLast;
                    EndMapIter emi=mEndMap.find(endid);
                    if (emi!=mEndMap.end())
                      {
                        EndIter end_iter=(*emi).second.mStrand;
                        (*end_iter).mShort = still_short;
                      }
                    endid.mContour = & (*(*best[1]).mIter);
                    endid.mFirstLast = (*best[1]).mFirstLast;
                    mEndMap.erase(endid);

                    endid.mContour = & (*(*best[1]).mIter);
                    endid.mFirstLast = 1-(*best[1]).mFirstLast;
                    emi=mEndMap.find(endid);
                    endid.mContour = & (*(*best[0]).mIter);
                    endid.mFirstLast = (*best[0]).mFirstLast;
                    EndMapIter emi2=mEndMap.find(endid);
                    assert(emi2!=mEndMap.end());

                    if (emi==mEndMap.end())
                      {
                        mEndMap.erase(emi2);
                      }
                    else
                      {
                        EndIter end_iter=(*emi).second.mStrand;
                        (*end_iter).mIter=(*best[0]).mIter;
                        (*end_iter).mFirstLast=(*best[0]).mFirstLast;
                        (*end_iter).mShort = still_short;
                        (*emi2).second = (*emi).second;
                        mEndMap.erase(emi);
                      }
                    contours.erase((*best[1]).mIter);
                  }
                // remove ends from star
                (*s).mEnds.erase(best[0]);
                (*s).mEnds.erase(best[1]);
#ifndef NDEBUG
                CheckEndMap();
#endif
              }
          }
      }

    bool PrivateImpl ::
    CheckCorners(int numcorners, CornerNbhd *corners, ScalarType epsilon)
      {
        bool ret = false;
        for (StarIter s=mStars.begin(); s!=mStars.end(); ++s)
          if (((*s).mEnds.size() % 2) == 1)
            {
              DomainType2d midpt = (*s).mBounds[0];
              Matrix::Add(midpt, midpt, (*s).mBounds[1]);
              Matrix::Scale(midpt, .5);

              int bestcorner=0;
              ScalarType bestdist=Matrix::Length(midpt, corners[0].mLocation);
              for (int c=1; c<numcorners; ++c)
                {
                  ScalarType dist=Matrix::Length(midpt, corners[c].mLocation);
                  if (dist<bestdist)
                    {
                      bestcorner=c;
                      bestdist=dist;
                    }
                }
              if (bestdist<epsilon && (*s).mSide!=corners[bestcorner].mSide)
                {
                  // change all ends in (*s).mEnds to corners[bestcorner].mSide
                  for (EndIter e=(*s).mEnds.begin(); e!=(*s).mEnds.end(); ++e)
                    {
                      assert((*(*e).mIter).mEndingSides[(*e).mFirstLast]==
                        (*s).mSide);
                      (*(*e).mIter).mEndingSides[(*e).mFirstLast]=
                        corners[bestcorner].mSide;
                    }
                  ret=true;
                }
            }
        return ret;
      }
    
    void PrivateImpl ::
    DoMerging(ContourList &dest, ScalarType epsilon, int numtomerge, 
              int *tomerge, int numcorners, CornerNbhd *corners)
      {
        PrivateImpl impl;
        impl.CreateStars(dest, 0, epsilon, numtomerge, tomerge);
#ifndef NDEBUG
        impl.CheckEndMap();
#endif
        if (numcorners && impl.CheckCorners(numcorners, corners, 3*epsilon))
          {
            impl.mStars.clear();
            impl.mEndMap.clear();
            impl.CreateStars(dest, 0, epsilon, numtomerge, tomerge);
#ifndef NDEBUG
            impl.CheckEndMap();
#endif
          }

        impl.ConnectStars(dest);
      }

    
    void PrivateImpl ::
    DoConnecting(ContourList &dest, ContourList &source, ScalarType epsilon, 
                  int numtomerge, int *tomerge)
      {
        PrivateImpl impl;
        
        impl.CreateStars(dest, 0, epsilon, numtomerge, tomerge);
        impl.CreateStars(source, 1, epsilon, numtomerge, tomerge);
        impl.ConnectStarValenceTwo(dest, source);
        // move remaining source contours to dest
        dest.splice(dest.end(), source);
      }

    
    void MergeContours(std::list<Root2D::Contour > &dest, 
              std::list<Root2D::Contour > &source, 
              ScalarType epsilon, int numtomerge, int *tomerge,
              int numcorners, CornerNbhd *corners)
      {
        dest.splice(dest.end(), source);
        PrivateImpl 
          :: DoMerging(dest, epsilon, numtomerge, tomerge, numcorners, corners);
      }

    
    void MergeContours(std::list<Root2D::Contour > &dest, 
              ScalarType epsilon, int numtomerge, int *tomerge)
      {
        PrivateImpl 
          :: DoMerging(dest, epsilon, numtomerge, tomerge, 0, 0);
      }

    
    void ConnectContours(std::list<Root2D::Contour > &dest, 
              std::list<Root2D::Contour > &source, 
              ScalarType epsilon, int numtomerge, int *tomerge)
      {
        PrivateImpl 
          :: DoConnecting(dest, source, epsilon, numtomerge, tomerge);
      }  
  }